import json
from six.moves.urllib.parse import parse_qs
from xml.dom.minidom import parseString


class XeroException(Exception):
    def __init__(self, response, msg=None):
        self.response = response
        super(XeroException, self).__init__(msg)


class XeroNotVerified(Exception):
    # Credentials haven't been verified
    pass


class XeroAccessDenied(Exception):
    # Access was denied
    pass


class XeroTenantIdNotSet(Exception):
    # Tenant Id must be set when using API with OAuth2 credentials
    pass


class XeroBadRequest(XeroException):
    # HTTP 400: Bad Request
    def __init__(self, response):
        if response.headers["content-type"].startswith("application/json"):
            data = json.loads(response.text)
            msg = "%s: %s" % (data["Type"], data["Message"])
            self.errors = [
                err["Message"]
                for elem in data.get("Elements", [])
                for err in elem.get("ValidationErrors", [])
            ]
            if len(self.errors) > 0:
                self.problem = self.errors[0]
                if len(self.errors) > 1:
                    msg += " (%s, and %s other issues)" % (
                        self.problem,
                        len(self.errors),
                    )
                else:
                    msg += " (%s)" % self.problem
            else:
                self.problem = None
            super(XeroBadRequest, self).__init__(response, msg=msg)

        elif response.headers["content-type"].startswith("text/html"):
            payload = parse_qs(response.text)
            self.errors = [payload["oauth_problem"][0]]
            self.problem = self.errors[0]
            super(XeroBadRequest, self).__init__(
                response, payload["oauth_problem_advice"][0]
            )

        else:
            # Extract the messages from the text.
            # parseString takes byte content, not unicode.
            dom = parseString(response.text.encode(response.encoding))
            messages = dom.getElementsByTagName("Message")

            msg = messages[0].childNodes[0].data
            self.errors = [m.childNodes[0].data for m in messages[1:]]
            self.problem = self.errors[0]
            super(XeroBadRequest, self).__init__(response, msg)


class XeroUnauthorized(XeroException):
    # HTTP 401: Unauthorized
    def __init__(self, response):
        payload = parse_qs(response.text)
        self.errors = [payload["oauth_problem"][0]]
        self.problem = self.errors[0]
        super(XeroUnauthorized, self).__init__(
            response, payload["oauth_problem_advice"][0]
        )


class XeroForbidden(XeroException):
    # HTTP 403: Forbidden
    def __init__(self, response):
        super(XeroForbidden, self).__init__(response, response.text)


class XeroNotFound(XeroException):
    # HTTP 404: Not Found
    def __init__(self, response):
        super(XeroNotFound, self).__init__(response, response.text)


class XeroUnsupportedMediaType(XeroException):
    # HTTP 415: UnsupportedMediaType
    def __init__(self, response):
        super(XeroUnsupportedMediaType, self).__init__(response, response.text)


class XeroInternalError(XeroException):
    # HTTP 500: Internal Error
    def __init__(self, response):
        super(XeroInternalError, self).__init__(response, response.text)


class XeroNotImplemented(XeroException):
    # HTTP 501
    def __init__(self, response):
        # Extract the useful error message from the text.
        # parseString takes byte content, not unicode.
        dom = parseString(response.text.encode(response.encoding))
        messages = dom.getElementsByTagName("Message")

        msg = messages[0].childNodes[0].data
        super(XeroNotImplemented, self).__init__(response, msg)


class XeroRateLimitExceeded(XeroException):
    # HTTP 503 - Rate limit exceeded
    def __init__(self, response, payload):
        try:
            self.errors = [payload["oauth_problem"][0]]
        except KeyError:
            return super(XeroRateLimitExceeded, self).__init__(response, response.text)
        self.problem = self.errors[0]
        super(XeroRateLimitExceeded, self).__init__(
            response, payload["oauth_problem_advice"][0]
        )


class XeroNotAvailable(XeroException):
    # HTTP 503 - Not available
    def __init__(self, response):
        super(XeroNotAvailable, self).__init__(response, response.text)


class XeroExceptionUnknown(XeroException):
    # Any other exception.
    pass



from __future__ import unicode_literals

import os
import requests
from six.moves.urllib.parse import parse_qs

from .constants import XERO_PROJECTS_URL
from .exceptions import (
    XeroBadRequest,
    XeroExceptionUnknown,
    XeroForbidden,
    XeroInternalError,
    XeroNotAvailable,
    XeroNotFound,
    XeroNotImplemented,
    XeroRateLimitExceeded,
    XeroUnauthorized,
    XeroUnsupportedMediaType,
)


class ProjectManager(object):
    DECORATED_METHODS = (
        "get",
        "all",
        "create",
        "delete",
        "get_tasks",
        "get_time",
        "set_status",
    )

    def __init__(self, name, credentials):
        self.credentials = credentials
        self.name = name
        self.base_url = credentials.base_url + XERO_PROJECTS_URL

        for method_name in self.DECORATED_METHODS:
            method = getattr(self, "_%s" % method_name)
            setattr(self, method_name, self._get_data(method))

    def _get_results(self, data):
        response = data["Response"]
        if self.name in response:
            result = response[self.name]
        elif "Attachments" in response:
            result = response["Attachments"]
        else:
            return None

        if isinstance(result, tuple) or isinstance(result, list):
            return result

        if isinstance(result, dict) and self.singular in result:
            return result[self.singular]

    def _get_data(self, func):
        """ This is the decorator for our DECORATED_METHODS.
        Each of the decorated methods must return:
            uri, params, method, body, headers, singleobject
        """

        def wrapper(*args, **kwargs):
            uri, params, method, body, headers, singleobject, files = func(
                *args, **kwargs
            )

            response = getattr(requests, method)(
                uri,
                data=body,
                headers=headers,
                auth=self.credentials.oauth,
                params=params,
                files=files,
            )

            if response.status_code == 200 or response.status_code == 201:
                if response.headers["content-type"].startswith("application/json"):
                    return response.json()
                else:
                    # return a byte string without doing any Unicode conversions
                    return response.content

            # Delete will return a response code of 204 - No Content
            elif response.status_code == 204:
                return "Deleted"

            elif response.status_code == 400:
                raise XeroBadRequest(response)

            elif response.status_code == 401:
                raise XeroUnauthorized(response)

            elif response.status_code == 403:
                raise XeroForbidden(response)

            elif response.status_code == 404:
                raise XeroNotFound(response)

            elif response.status_code == 415:
                raise XeroUnsupportedMediaType(response)

            elif response.status_code == 500:
                raise XeroInternalError(response)

            elif response.status_code == 501:
                raise XeroNotImplemented(response)

            elif response.status_code == 503:
                # Two 503 responses are possible. Rate limit errors
                # return encoded content; offline errors don't.
                # If you parse the response text and there's nothing
                # encoded, it must be a not-available error.
                payload = parse_qs(response.text)
                if payload:
                    raise XeroRateLimitExceeded(response, payload)
                else:
                    raise XeroNotAvailable(response)
            else:
                raise XeroExceptionUnknown(response)

        return wrapper

    def _get(self, id, headers=None):
        uri = "/".join([self.base_url, self.name, id])
        return uri, {}, "get", None, headers, True, None

    def _get_tasks(self, projectId):
        """Retrieve the list of tasks contained in a project"""
        uri = "/".join([self.base_url, self.name, projectId, "Tasks"])
        return uri, {}, "get", None, None, False, None

    def _get_time(self, projectId):
        """Retrieve the list of times contained in a project"""
        uri = "/".join([self.base_url, self.name, projectId, "Time"])
        return uri, {}, "get", None, None, False, None

    def _set_status(self, projectId, data):
        uri = "/".join([self.base_url, self.name, projectId])
        body = data
        return uri, {}, "patch", body, None, False, None

    def create_or_save(self, data, method="post", headers=None, summarize_errors=True):
        if "Id" not in data:
            uri = "/".join([self.base_url, self.name])
        else:
            uri = "/".join([self.base_url, self.name, data["Id"]])
        body = data
        if summarize_errors:
            params = {}
        else:
            params = {"summarizeErrors": "false"}
        return uri, params, method, body, headers, False, None

    def _create(self, data):
        return self.create_or_save(data, method="post")

    def _delete(self, id):
        uri = "/".join([self.base_url, self.name, id])
        return uri, {}, "delete", None, None, False, None

    def _all(self):
        uri = "/".join([self.base_url, self.name])
        return uri, {}, "get", None, None, False, None

    def filename(self, path):
        head, tail = os.path.split(path)
        return tail or os.path.basename(head)



from __future__ import unicode_literals

import datetime
import unittest

import xero.utils

""" Tests of the utils module. """


class UtilsTest(unittest.TestCase):
    """ Test of the utils module.
    """

    def test_json_hook(self):
        """ Tests the json hook used in Manager._parse_api_response, and the
            call it makes to parse_date.
        """
        # The hook parses dates
        example_input = {
            "date": "/Date(1426849200000+1300)/",
        }

        self.assertEqual(
            xero.utils.json_load_object_hook(example_input),
            {"date": datetime.datetime(2015, 3, 21, 0, 0)},
        )

        # In both format styles
        example_input = {
            "date": "2015-04-29T00:00:00",
        }

        self.assertEqual(
            xero.utils.json_load_object_hook(example_input),
            {"date": datetime.date(2015, 4, 29)},
        )

        # And in unicode
        example_input = {
            "date": "2015-04-29T10:21:03",
        }

        self.assertEqual(
            xero.utils.json_load_object_hook(example_input),
            {"date": datetime.datetime(2015, 4, 29, 10, 21, 3)},
        )

        # Not a string type
        self.assertEqual(xero.utils.json_load_object_hook({"date": 6}), {"date": 6})

        self.assertEqual(
            xero.utils.json_load_object_hook({"date": None}), {"date": None}
        )

        # Weird Date output from Xero
        example_input = {
            "date": "/Date(0+0000)/",
        }

        self.assertEqual(
            xero.utils.json_load_object_hook(example_input), {"date": "/Date(0+0000)/"}
        )

    def test_parse_date(self):
        """ Tests of the parse_date input formats.
        """
        # 07/05/2015 00:00:00 +12 (06/05/2015 12:00:00 GMT/UTC)
        self.assertEqual(
            xero.utils.parse_date("/Date(1430913600000+1200)/"),
            datetime.datetime(2015, 5, 7, 0, 0),
        )

        # 16/09/2008 10:28:51.5 +12 (15/09/2008 22:25:51.5 GMT/UTC)
        self.assertEqual(
            xero.utils.parse_date("/Date(1221517731500+1200)/"),
            datetime.datetime(2008, 9, 16, 10, 28, 51, 500000),
        )

        # 10/08/2015 10:55:33 GMT/UTC
        self.assertEqual(
            xero.utils.parse_date("/Date(1439204133355)/"),
            datetime.datetime(2015, 8, 10, 10, 55, 33, 355000),
        )

        # 29/04/2015 00:00:00
        self.assertEqual(
            xero.utils.parse_date("2015-04-29T00:00:00"), datetime.date(2015, 4, 29)
        )

        # 29/04/2015 10:21:03
        self.assertEqual(
            xero.utils.parse_date("2015-04-29T10:21:03"),
            datetime.datetime(2015, 4, 29, 10, 21, 3),
        )

        # Not a date
        self.assertEqual(xero.utils.parse_date("not a date"), None)

        # Weird Date output from Xero
        self.assertEqual(xero.utils.parse_date("/Date(0+0000)/"), None)



from __future__ import unicode_literals

import datetime
import re
import six
import unittest
from collections import defaultdict
from mock import Mock
from xml.dom.minidom import parseString

from xero.manager import Manager


class ManagerTest(unittest.TestCase):
    def assertXMLEqual(self, xml1, xml2, message=""):
        def to_str(s):
            return s.decode("utf-8") if six.PY3 and isinstance(s, bytes) else str(s)

        def clean_xml(xml):
            xml = "<root>%s</root>" % to_str(xml)
            return str(re.sub(">\n *<", "><", parseString(xml).toxml()))

        def xml_to_dict(xml):
            nodes = re.findall("(<([^>]*)>(.*?)</\\2>)", xml)
            if len(nodes) == 0:
                return xml
            d = defaultdict(list)
            for node in nodes:
                d[node[1]].append(xml_to_dict(node[2]))
            return d

        cleaned = map(clean_xml, (xml1, xml2))
        d1, d2 = tuple(map(xml_to_dict, cleaned))

        self.assertEqual(d1, d2, message)

    def test_serializer(self):
        credentials = Mock(base_url="")
        manager = Manager("contacts", credentials)

        example_invoice_input = {
            "Date": datetime.datetime(2015, 6, 6, 16, 25, 2, 711109),
            "Reference": "ABAS 123",
            "LineItems": [
                {"Description": "Example description only"},
                {
                    "UnitAmount": "0.0000",
                    "Quantity": 1,
                    "AccountCode": "200",
                    "Description": "Example line item 2",
                    "TaxType": "OUTPUT",
                },
                {
                    "UnitAmount": "231.0000",
                    "Quantity": 1,
                    "AccountCode": "200",
                    "Description": "Example line item 3",
                    "TaxType": "OUTPUT",
                },
            ],
            "Status": "DRAFT",
            "Type": "ACCREC",
            "DueDate": datetime.datetime(2015, 7, 6, 16, 25, 2, 711136),
            "LineAmountTypes": "Exclusive",
            "Contact": {"Name": "Basket Case"},
        }
        resultant_xml = manager._prepare_data_for_save(example_invoice_input)
        resultant_xml = "<Invoice>%s</Invoice>" % resultant_xml

        expected_xml = """
        <Invoice>
          <Status>DRAFT</Status>
          <Contact>
            <Name>Basket Case</Name>
          </Contact>
          <Reference>ABAS 123</Reference>
          <Date>2015-06-06T16:25:02</Date>
          <LineAmountTypes>Exclusive</LineAmountTypes>
          <LineItems>
            <LineItem>
              <Description>Example description only</Description>
            </LineItem>
            <LineItem>
              <TaxType>OUTPUT</TaxType>
              <AccountCode>200</AccountCode>
              <UnitAmount>0.0000</UnitAmount>
              <Description>Example line item 2</Description>
              <Quantity>1</Quantity>
            </LineItem>
            <LineItem>
              <TaxType>OUTPUT</TaxType>
              <AccountCode>200</AccountCode>
              <UnitAmount>231.0000</UnitAmount>
              <Description>Example line item 3</Description>
              <Quantity>1</Quantity>
            </LineItem>
          </LineItems>
          <Type>ACCREC</Type>
          <DueDate>2015-07-06T16:25:02</DueDate>
        </Invoice>
        """

        self.assertXMLEqual(
            resultant_xml, expected_xml,
        )

    def test_serializer_phones_addresses(self):
        credentials = Mock(base_url="")
        manager = Manager("contacts", credentials)

        example_contact_input = {
            "ContactID": "565acaa9-e7f3-4fbf-80c3-16b081ddae10",
            "ContactStatus": "ACTIVE",
            "Name": "Southside Office Supplies",
            "Addresses": [{"AddressType": "POBOX"}, {"AddressType": "STREET"}],
            "Phones": [
                {"PhoneType": "DDI"},
                {"PhoneType": "DEFAULT"},
                {"PhoneType": "FAX"},
                {"PhoneType": "MOBILE"},
            ],
            "UpdatedDateUTC": datetime.datetime(2015, 9, 18, 5, 6, 56, 893),
            "IsSupplier": False,
            "IsCustomer": False,
            "HasAttachments": False,
        }
        resultant_xml = manager._prepare_data_for_save(example_contact_input)
        resultant_xml = "<Contact>%s</Contact>" % resultant_xml

        expected_xml = """
        <Contact>
          <ContactID>565acaa9-e7f3-4fbf-80c3-16b081ddae10</ContactID>
          <Name>Southside Office Supplies</Name>
          <HasAttachments>false</HasAttachments>
          <Phones>
            <Phone>
              <PhoneType>DDI</PhoneType>
            </Phone>
            <Phone>
              <PhoneType>DEFAULT</PhoneType>
            </Phone>
            <Phone>
              <PhoneType>FAX</PhoneType>
            </Phone>
            <Phone>
              <PhoneType>MOBILE</PhoneType>
            </Phone>
          </Phones>
          <IsCustomer>false</IsCustomer>
          <Addresses>
            <Address>
              <AddressType>POBOX</AddressType>
            </Address>
            <Address>
              <AddressType>STREET</AddressType>
            </Address>
          </Addresses>
          <IsSupplier>false</IsSupplier>
          <ContactStatus>ACTIVE</ContactStatus>
        </Contact>
        """

        self.assertXMLEqual(
            resultant_xml, expected_xml, "Resultant XML does not match expected."
        )

    def test_serializer_nested_singular(self):
        credentials = Mock(base_url="")
        manager = Manager("contacts", credentials)

        example_invoice_input = {
            "Date": datetime.datetime(2015, 6, 6, 16, 25, 2, 711109),
            "Reference": "ABAS 123",
            "LineItems": [{"Description": "Example description only"}],
            "Status": "DRAFT",
            "Type": "ACCREC",
            "DueDate": datetime.datetime(2015, 7, 6, 16, 25, 2, 711136),
            "LineAmountTypes": "Exclusive",
            "Contact": {"Name": "Basket Case"},
        }
        resultant_xml = manager._prepare_data_for_save(example_invoice_input)

        expected_xml = """
            <Status>DRAFT</Status>
            <Contact><Name>Basket Case</Name></Contact>
            <Reference>ABAS 123</Reference>
            <Date>2015-06-06T16:25:02</Date>
            <LineAmountTypes>Exclusive</LineAmountTypes>
            <LineItems>
              <LineItem>
                <Description>Example description only</Description>
              </LineItem>
            </LineItems>
            <Type>ACCREC</Type>
            <DueDate>2015-07-06T16:25:02</DueDate>
        """

        self.assertXMLEqual(
            resultant_xml, expected_xml,
        )

    def test_filter(self):
        """The filter function should correctly handle various arguments"""
        credentials = Mock(base_url="")
        manager = Manager("contacts", credentials)

        uri, params, method, body, headers, singleobject = manager._filter(
            order="LastName",
            page=2,
            offset=5,
            since=datetime.datetime(2014, 8, 10, 15, 14, 46),
            Name="John",
        )

        self.assertEqual(method, "get")
        self.assertFalse(singleobject)

        expected_params = {
            "order": "LastName",
            "page": 2,
            "offset": 5,
            "where": 'Name=="John"',
        }
        self.assertEqual(params, expected_params)

        expected_headers = {"If-Modified-Since": "Sun, 10 Aug 2014 15:14:46 GMT"}
        self.assertEqual(headers, expected_headers)

        # Also make sure an empty call runs ok
        uri, params, method, body, headers, singleobject = manager._filter()
        self.assertEqual(params, {})
        self.assertIsNone(headers)

        manager = Manager("invoices", credentials)
        uri, params, method, body, headers, singleobject = manager._filter(
            **{"Contact.ContactID": "3e776c4b-ea9e-4bb1-96be-6b0c7a71a37f"}
        )

        self.assertEqual(
            params,
            {
                "where": 'Contact.ContactID==Guid("3e776c4b-ea9e-4bb1-96be-6b0c7a71a37f")'
            },
        )

        (uri, params, method, body, headers, singleobject) = manager._filter(
            **{"AmountPaid": 0.0}
        )

        self.assertEqual(params, {"where": 'AmountPaid=="0.0"'})

    def test_rawfilter(self):
        """The filter function should correctly handle various arguments"""
        credentials = Mock(base_url="")
        manager = Manager("invoices", credentials)
        uri, params, method, body, headers, singleobject = manager._filter(
            Status="VOIDED", raw='Name.ToLower()=="test contact"'
        )
        self.assertEqual(
            params, {"where": 'Name.ToLower()=="test contact"&&Status=="VOIDED"'}
        )

    def test_boolean_filter(self):
        """The filter function should correctly handle various arguments"""
        credentials = Mock(base_url="")
        manager = Manager("invoices", credentials)
        uri, params, method, body, headers, singleobject = manager._filter(
            CanApplyToRevenue=True
        )
        self.assertEqual(params, {"where": "CanApplyToRevenue==true"})

    def test_magnitude_filters(self):
        """The filter function should correctlu handle date arguments and gt, lt operators"""
        credentials = Mock(base_url="")

        manager = Manager("invoices", credentials)
        uri, params, method, body, headers, singleobject = manager._filter(
            **{"Date__gt": datetime.datetime(2007, 12, 6)}
        )

        self.assertEqual(params, {"where": "Date>DateTime(2007,12,6)"})

        manager = Manager("invoices", credentials)
        uri, params, method, body, headers, singleobject = manager._filter(
            **{"Date__lte": datetime.datetime(2007, 12, 6)}
        )

        self.assertEqual(params, {"where": "Date<=DateTime(2007,12,6)"})

    def test_unit4dps(self):
        """The manager should add a query param of unitdp iff enabled"""

        credentials = Mock(base_url="")

        # test 4dps is disabled by default
        manager = Manager("contacts", credentials)
        uri, params, method, body, headers, singleobject = manager._filter()
        self.assertEqual(params, {}, "test 4dps not enabled by default")

        # test 4dps is enabled by default
        manager = Manager("contacts", credentials, unit_price_4dps=True)
        uri, params, method, body, headers, singleobject = manager._filter()
        self.assertEqual(params, {"unitdp": 4}, "test 4dps can be enabled explicitly")

        # test 4dps can be disable explicitly
        manager = Manager("contacts", credentials, unit_price_4dps=False)
        uri, params, method, body, headers, singleobject = manager._filter()
        self.assertEqual(params, {}, "test 4dps can be disabled explicitly")

    def test_get_params(self):
        """The 'get' methods should pass GET parameters if provided.
        """

        credentials = Mock(base_url="")
        manager = Manager("reports", credentials)

        # test no parameters or headers sent by default
        uri, params, method, body, headers, singleobject = manager._get("ProfitAndLoss")
        self.assertEqual(params, {}, "test params not sent by default")

        # test params can be provided
        passed_params = {
            "fromDate": "2015-01-01",
            "toDate": "2015-01-15",
        }
        uri, params, method, body, headers, singleobject = manager._get(
            "ProfitAndLoss", params=passed_params
        )
        self.assertEqual(params, passed_params, "test params can be set")

        # test params respect, but can override, existing configuration
        manager = Manager("reports", credentials, unit_price_4dps=True)
        uri, params, method, body, headers, singleobject = manager._get(
            "ProfitAndLoss", params=passed_params
        )
        self.assertEqual(
            params,
            {"fromDate": "2015-01-01", "toDate": "2015-01-15", "unitdp": 4},
            "test params respects existing values",
        )

    def test_user_agent_inheritance(self):
        """The user_agent should be inherited from the provided credentials when not set explicitly.
        """

        # Default used when no user_agent set on manager and credentials has nothing to offer.
        credentials = Mock(base_url="", user_agent=None)
        manager = Manager("reports", credentials)
        self.assertTrue(manager.user_agent.startswith("pyxero/"))

        # Taken from credentials when no user_agent set on manager.
        credentials = Mock(base_url="", user_agent="MY_COMPANY-MY_CONSUMER_KEY")
        manager = Manager("reports", credentials)
        self.assertEqual(manager.user_agent, "MY_COMPANY-MY_CONSUMER_KEY")

        # Manager's user_agent used when explicitly set.
        credentials = Mock(base_url="", user_agent="MY_COMPANY-MY_CONSUMER_KEY")
        manager = Manager("reports", credentials, user_agent="DemoCompany-1234567890")
        self.assertEqual(manager.user_agent, "DemoCompany-1234567890")



from __future__ import unicode_literals

from .filesmanager import FilesManager
from .manager import Manager
from .payrollmanager import PayrollManager
from .projectmanager import ProjectManager


class Xero(object):
    """An ORM-like interface to the Xero API"""

    OBJECT_LIST = (
        "Attachments",
        "Accounts",
        "BankTransactions",
        "BankTransfers",
        "BrandingThemes",
        "BatchPayments",
        "ContactGroups",
        "Contacts",
        "CreditNotes",
        "Currencies",
        "Employees",
        "ExpenseClaims",
        "Invoices",
        "Items",
        "Journals",
        "ManualJournals",
        "Organisations",
        "Overpayments",
        "Payments",
        "Prepayments",
        "PurchaseOrders",
        "Receipts",
        "RepeatingInvoices",
        "Reports",
        "TaxRates",
        "TrackingCategories",
        "Users",
        "Quotes",
    )

    def __init__(self, credentials, unit_price_4dps=False, user_agent=None):
        # Iterate through the list of objects we support, for
        # each of them create an attribute on our self that is
        # the lowercase name of the object and attach it to an
        # instance of a Manager object to operate on it
        for name in self.OBJECT_LIST:
            setattr(
                self,
                name.lower(),
                Manager(name, credentials, unit_price_4dps, user_agent),
            )

        setattr(self, "filesAPI", Files(credentials))
        setattr(self, "payrollAPI", Payroll(credentials, unit_price_4dps, user_agent))
        setattr(self, "projectsAPI", Project(credentials))


class Files(object):
    """An ORM-like interface to the Xero Files API"""

    OBJECT_LIST = (
        "Associations",
        "Files",
        "Folders",
        "Inbox",
    )

    def __init__(self, credentials):
        # Iterate through the list of objects we support, for
        # each of them create an attribute on our self that is
        # the lowercase name of the object and attach it to an
        # instance of a Manager object to operate on it
        for name in self.OBJECT_LIST:
            setattr(self, name.lower(), FilesManager(name, credentials))


class Payroll(object):
    """An ORM-like interface to the Xero Payroll API"""

    OBJECT_LIST = (
        "Employees",
        "SuperFunds",
        "Timesheets",
        "PayItems",
        "PayRuns",
        "PayrollCalendars",
        "Payslip",
        "LeaveApplications",
    )

    def __init__(self, credentials, unit_price_4dps=False, user_agent=None):
        for name in self.OBJECT_LIST:
            setattr(
                self,
                name.lower(),
                PayrollManager(name, credentials, unit_price_4dps, user_agent),
            )


class Project(object):
    """An ORM-like interface to the Xero Projects API"""

    OBJECT_LIST = (
        "Projects",
        "Projectsusers",
        "Tasks",
        "Time",
    )

    def __init__(self, credentials):
        # Iterate through the list of objects we support, for
        # each of them create an attribute on our self that is
        # the lowercase name of the object and attach it to an
        # instance of a Manager object to operate on it
        for name in self.OBJECT_LIST:
            setattr(self, name.lower(), ProjectManager(name, credentials))



from __future__ import unicode_literals

import datetime
import requests
from six.moves.urllib.parse import parse_qs, urlencode

from oauthlib.oauth1 import SIGNATURE_HMAC, SIGNATURE_RSA, SIGNATURE_TYPE_AUTH_HEADER
from requests_oauthlib import OAuth1, OAuth2, OAuth2Session

from .constants import (
    ACCESS_TOKEN_URL,
    AUTHORIZE_URL,
    REQUEST_TOKEN_URL,
    XERO_BASE_URL,
    XERO_OAUTH2_AUTHORIZE_URL,
    XERO_OAUTH2_CONNECTIONS_URL,
    XERO_OAUTH2_TOKEN_URL,
    XeroScopes,
)
from .exceptions import (
    XeroAccessDenied,
    XeroBadRequest,
    XeroException,
    XeroExceptionUnknown,
    XeroForbidden,
    XeroInternalError,
    XeroNotAvailable,
    XeroNotFound,
    XeroNotImplemented,
    XeroNotVerified,
    XeroRateLimitExceeded,
    XeroUnauthorized,
)
from .utils import resolve_user_agent

OAUTH_EXPIRY_SECONDS = 3600  # Default unless a response reports differently

DEFAULT_SCOPE = [
    XeroScopes.OFFLINE_ACCESS,
    XeroScopes.ACCOUNTING_TRANSACTIONS_READ,
    XeroScopes.ACCOUNTING_CONTACTS_READ,
]


class PrivateCredentials:
    """An object wrapping the 2-step OAuth process for Private Xero API access.

    Usage:

     1) Construct a PrivateCredentials() instance:

        >>> from xero.auth import PrivateCredentials
        >>> credentials = PrivateCredentials(<consumer_key>, <rsa_key>)

        rsa_key should be a multi-line string, starting with:

            -----BEGIN RSA PRIVATE KEY-----\n

     2) Use the credentials:

        >>> from xero import Xero
        >>> xero = Xero(credentials)
        >>> xero.contacts.all()
        ...
    """

    def __init__(self, consumer_key, rsa_key, api_url=XERO_BASE_URL):
        self.consumer_key = consumer_key
        self.rsa_key = rsa_key

        self.base_url = api_url

        # Private API uses consumer key as the OAuth token.
        self.oauth_token = consumer_key

        self.oauth = OAuth1(
            self.consumer_key,
            resource_owner_key=self.oauth_token,
            rsa_key=self.rsa_key,
            signature_method=SIGNATURE_RSA,
            signature_type=SIGNATURE_TYPE_AUTH_HEADER,
        )


class PublicCredentials:
    """An object wrapping the 3-step OAuth process for Public Xero API access.

    Usage:

     1) Construct a PublicCredentials() instance:

        >>> from xero import PublicCredentials
        >>> credentials = PublicCredentials(<consumer_key>, <consumer_secret>)

     2) Visit the authentication URL:

        >>> credentials.url

        If a callback URI was provided (e.g., https://example.com/oauth),
        the user will be redirected to a URL of the form:

        https://example.com/oauth?oauth_token=<token>&oauth_verifier=<verifier>&org=<organization ID>

        from which the verifier can be extracted. If no callback URI is
        provided, the verifier will be shown on the screen, and must be
        manually entered by the user.

     3) Verify the instance:

        >>> credentials.verify(<verifier string>)

     4) Use the credentials.

        >>> from xero import Xero
        >>> xero = Xero(credentials)
        >>> xero.contacts.all()
        ...
    """

    def __init__(
        self,
        consumer_key,
        consumer_secret,
        callback_uri=None,
        verified=False,
        oauth_token=None,
        oauth_token_secret=None,
        oauth_expires_at=None,
        oauth_authorization_expires_at=None,
        scope=None,
        user_agent=None,
        api_url=XERO_BASE_URL,
    ):
        """Construct the auth instance.

        Must provide the consumer key and secret.
        A callback URL may be provided as an option. If provided, the
        Xero verification process will redirect to that URL when

        """
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.callback_uri = callback_uri
        self.verified = verified
        self._oauth = None
        self.oauth_expires_at = oauth_expires_at
        self.oauth_authorization_expires_at = oauth_authorization_expires_at
        self.scope = scope
        self.user_agent = resolve_user_agent(user_agent)

        self.base_url = api_url
        self._signature_method = SIGNATURE_HMAC

        # These are not strictly used by Public Credentials, but
        # are reserved for use by other credentials (i.e. Partner)
        self.rsa_key = None
        self.oauth_session_handle = None

        self._init_credentials(oauth_token, oauth_token_secret)

    def _init_credentials(self, oauth_token, oauth_token_secret):
        "Depending on the state passed in, get self._oauth up and running"
        if oauth_token and oauth_token_secret:
            if self.verified:
                # If provided, this is a fully verified set of
                # credentials. Store the oauth_token and secret
                # and initialize OAuth around those
                self._init_oauth(oauth_token, oauth_token_secret)

            else:
                # If provided, we are reconstructing an initalized
                # (but non-verified) set of public credentials.
                self.oauth_token = oauth_token
                self.oauth_token_secret = oauth_token_secret

        else:
            # This is a brand new set of credentials - we need to generate
            # an oauth token so it's available for the url property.
            oauth = OAuth1(
                self.consumer_key,
                client_secret=self.consumer_secret,
                callback_uri=self.callback_uri,
                rsa_key=self.rsa_key,
                signature_method=self._signature_method,
            )

            url = self.base_url + REQUEST_TOKEN_URL
            headers = {"User-Agent": self.user_agent}
            response = requests.post(url=url, headers=headers, auth=oauth)
            self._process_oauth_response(response)

    def _init_oauth(self, oauth_token, oauth_token_secret):
        "Store and initialize a verified set of OAuth credentials"
        self.oauth_token = oauth_token
        self.oauth_token_secret = oauth_token_secret

        self._oauth = OAuth1(
            self.consumer_key,
            client_secret=self.consumer_secret,
            resource_owner_key=self.oauth_token,
            resource_owner_secret=self.oauth_token_secret,
            rsa_key=self.rsa_key,
            signature_method=self._signature_method,
        )

    def _process_oauth_response(self, response):
        "Extracts the fields from an oauth response"
        if response.status_code == 200:
            credentials = parse_qs(response.text)

            # Initialize the oauth credentials
            self._init_oauth(
                credentials.get("oauth_token")[0],
                credentials.get("oauth_token_secret")[0],
            )

            # If tokens are refreshable, we'll get a session handle
            self.oauth_session_handle = credentials.get("oauth_session_handle", [None])[
                0
            ]

            # Calculate token/auth expiry
            oauth_expires_in = credentials.get(
                "oauth_expires_in", [OAUTH_EXPIRY_SECONDS]
            )[0]
            oauth_authorisation_expires_in = credentials.get(
                "oauth_authorization_expires_in", [OAUTH_EXPIRY_SECONDS]
            )[0]

            self.oauth_expires_at = datetime.datetime.now() + datetime.timedelta(
                seconds=int(oauth_expires_in)
            )
            self.oauth_authorization_expires_at = datetime.datetime.now() + datetime.timedelta(
                seconds=int(oauth_authorisation_expires_in)
            )
        else:
            self._handle_error_response(response)

    def _handle_error_response(self, response):
        if response.status_code == 400:
            raise XeroBadRequest(response)

        elif response.status_code == 401:
            raise XeroUnauthorized(response)

        elif response.status_code == 403:
            raise XeroForbidden(response)

        elif response.status_code == 404:
            raise XeroNotFound(response)

        elif response.status_code == 500:
            raise XeroInternalError(response)

        elif response.status_code == 501:
            raise XeroNotImplemented(response)

        elif response.status_code == 503:
            # Two 503 responses are possible. Rate limit errors
            # return encoded content; offline errors don't.
            # If you parse the response text and there's nothing
            # encoded, it must be a not-available error.
            payload = parse_qs(response.text)
            if payload:
                raise XeroRateLimitExceeded(response, payload)
            else:
                raise XeroNotAvailable(response)
        else:
            raise XeroExceptionUnknown(response)

    @property
    def state(self):
        """Obtain the useful state of this credentials object so that
        we can reconstruct it independently.
        """
        return dict(
            (attr, getattr(self, attr))
            for attr in (
                "consumer_key",
                "consumer_secret",
                "callback_uri",
                "verified",
                "oauth_token",
                "oauth_token_secret",
                "oauth_session_handle",
                "oauth_expires_at",
                "oauth_authorization_expires_at",
                "scope",
            )
            if getattr(self, attr) is not None
        )

    def verify(self, verifier):
        "Verify an OAuth token"

        # Construct the credentials for the verification request
        oauth = OAuth1(
            self.consumer_key,
            client_secret=self.consumer_secret,
            resource_owner_key=self.oauth_token,
            resource_owner_secret=self.oauth_token_secret,
            verifier=verifier,
            rsa_key=self.rsa_key,
            signature_method=self._signature_method,
        )

        # Make the verification request, gettiung back an access token
        url = self.base_url + ACCESS_TOKEN_URL
        headers = {"User-Agent": self.user_agent}
        response = requests.post(url=url, headers=headers, auth=oauth)
        self._process_oauth_response(response)
        self.verified = True

    @property
    def url(self):
        "Returns the URL that can be visited to obtain a verifier code"
        # The authorize url is always api.xero.com
        query_string = {"oauth_token": self.oauth_token}

        if self.scope:
            query_string["scope"] = self.scope

        url = self.base_url + AUTHORIZE_URL + "?" + urlencode(query_string)
        return url

    @property
    def oauth(self):
        "Returns the requests-compatible OAuth object"
        if self._oauth is None:
            raise XeroNotVerified("OAuth credentials haven't been verified")
        return self._oauth

    def expired(self, now=None):
        if now is None:
            now = datetime.datetime.now()

        # Credentials states from older versions might not have
        # oauth_expires_at available
        if self.oauth_expires_at is None:
            raise XeroException(None, "Expiry time is not available")

        # Allow a bit of time for clock differences and round trip times
        # to prevent false negatives. If users want the precise expiry,
        # they can use self.oauth_expires_at
        CONSERVATIVE_SECONDS = 30

        return self.oauth_expires_at <= (
            now + datetime.timedelta(seconds=CONSERVATIVE_SECONDS)
        )


class PartnerCredentials(PublicCredentials):
    """An object wrapping the 3-step OAuth process for Partner Xero API access.

    Usage is very similar to Public Credentials with the following changes:

     1) You'll need to pass the private key for your RSA certificate.

        >>> rsa_key = "-----BEGIN RSA PRIVATE KEY----- ..."

     2) Once a token has expired, you can refresh it to get another 30 mins

        >>> credentials = PartnerCredentials(**state)
        >>> if credentials.expired():
                credentials.refresh()

     3) Authorization expiry and token expiry become different things.

        oauth_expires_at tells when the current token expires (~30 min window)

        oauth_authorization_expires_at tells when the overall access
        permissions expire (~10 year window)
    """

    def __init__(
        self,
        consumer_key,
        consumer_secret,
        rsa_key,
        callback_uri=None,
        verified=False,
        oauth_token=None,
        oauth_token_secret=None,
        oauth_expires_at=None,
        oauth_authorization_expires_at=None,
        oauth_session_handle=None,
        scope=None,
        user_agent=None,
        api_url=XERO_BASE_URL,
        **kwargs
    ):
        """Construct the auth instance.

        Must provide the consumer key and secret.
        A callback URL may be provided as an option. If provided, the
        Xero verification process will redirect to that URL when

        """
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.callback_uri = callback_uri
        self.verified = verified
        self._oauth = None
        self.oauth_expires_at = oauth_expires_at
        self.oauth_authorization_expires_at = oauth_authorization_expires_at
        self.scope = scope
        self.user_agent = resolve_user_agent(user_agent)

        self._signature_method = SIGNATURE_RSA
        self.base_url = api_url

        self.rsa_key = rsa_key
        self.oauth_session_handle = oauth_session_handle

        self._init_credentials(oauth_token, oauth_token_secret)

    def refresh(self):
        "Refresh an expired token"

        # Construct the credentials for the verification request
        oauth = OAuth1(
            self.consumer_key,
            client_secret=self.consumer_secret,
            resource_owner_key=self.oauth_token,
            resource_owner_secret=self.oauth_token_secret,
            rsa_key=self.rsa_key,
            signature_method=self._signature_method,
        )

        # Make the verification request, getting back an access token
        headers = {"User-Agent": self.user_agent}
        params = {"oauth_session_handle": self.oauth_session_handle}
        response = requests.post(
            url=self.base_url + ACCESS_TOKEN_URL,
            params=params,
            headers=headers,
            auth=oauth,
        )
        self._process_oauth_response(response)


class OAuth2Credentials(object):
    """An object wrapping the 3-step OAuth2.0 process for Xero API access.

        For detailed documentation see README.md.
    Usage:

     1) Construct an `OAuth2Credentials` instance:
        >>> credentials = OAuth2Credentials(client_id, client_secret,
        >>>                                 callback_uri=callback_uri, scope=scope)

     2) Generate a unique authentication URL and visit it:
        >>> credentials.generate_url()

        The user will be redirected to a URL in the form:
        https://example.com/oauth/xero/callback/?code=0123456789&scope=openid%20profile
        &state=87784234sdf5ds8ad546a8sd545ss6

     3) Verify the credentials using the full URL redirected to, including querystring:
        >>> credentials.verify(full_url_with_querystring)

     4) Use the credentials. It is usually necessary to set the tenant_id (Xero
        organisation id) to specify the organisation against which the queries should
        run:
        >>> from xero import Xero
        >>> credentials.set_default_tenant()
        >>> xero = Xero(credentials)
        >>> xero.contacts.all()
        ...

        To use a different organisation, set credentials.tenant_id:
        >>> tenants = credentials.get_tenants()
        >>> credentials.tenant_id = tenants[1]['tenantId']

     5) If a refresh token is available, it can be used to generate a new token:
        >>> if credentials.expired():
        >>>     credentials.refresh()

        Note that in order for tokens to be refreshable, Xero API requires
        `offline_access` to be included in the scope.
    """

    def __init__(
        self,
        client_id,
        client_secret,
        callback_uri=None,
        auth_state=None,
        auth_secret=None,
        token=None,
        scope=None,
        tenant_id=None,
        user_agent=None,
    ):
        from xero import __version__ as VERSION

        self.client_id = client_id
        self.client_secret = client_secret
        self.callback_uri = callback_uri
        self.auth_state = auth_state
        self.token = None
        self.tenant_id = tenant_id  # Used by BaseManager
        self._oauth = None
        self.scope = scope or DEFAULT_SCOPE[:]

        if user_agent is None:
            self.user_agent = (
                "pyxero/%s " % VERSION + requests.utils.default_user_agent()
            )
        else:
            self.user_agent = user_agent

        self.base_url = XERO_BASE_URL  # Used by BaseManager
        self._init_credentials(token, auth_secret)

    def _init_credentials(self, token, auth_secret):
        """
        Depending on the state passed in, get self._oauth up and running.
        """
        if token:
            self._init_oauth(token)
        elif auth_secret and self.auth_state:
            self.verify(auth_secret)

    def _init_oauth(self, token):
        """Set self._oauth for use by the xero client."""
        self.token = token
        if token:
            self._oauth = OAuth2(client_id=self.client_id, token=self.token)

    @property
    def state(self):
        """Obtain the useful state of this credentials object so that
        we can reconstruct it independently.
        """
        return dict(
            (attr, getattr(self, attr))
            for attr in (
                "client_id",
                "client_secret",
                "callback_uri",
                "auth_state",
                "token",
                "scope",
                "tenant_id",
                "user_agent",
            )
            if getattr(self, attr) is not None
        )

    def verify(self, auth_secret):
        """Verify and return OAuth2 token."""
        session = OAuth2Session(
            self.client_id,
            state=self.auth_state,
            scope=self.scope,
            redirect_uri=self.callback_uri,
        )
        try:
            token = session.fetch_token(
                XERO_OAUTH2_TOKEN_URL,
                client_secret=self.client_secret,
                authorization_response=auth_secret,
                headers=self.headers,
            )
        # Various different exceptions may be raised, so pass the exception
        # through as XeroAccessDenied
        except Exception as e:
            raise XeroAccessDenied(e)
        self._init_oauth(token)

    def generate_url(self):
        """Get the authorization url. This will also set `self.auth_state` to a
        random string if it has not already been set.
        """
        session = OAuth2Session(
            self.client_id, scope=self.scope, redirect_uri=self.callback_uri
        )
        url, self.auth_state = session.authorization_url(
            XERO_OAUTH2_AUTHORIZE_URL, state=self.auth_state
        )
        return url

    @property
    def oauth(self):
        """Return the requests-compatible OAuth object"""
        if self._oauth is None:
            raise XeroNotVerified("OAuth credentials haven't been verified")
        return self._oauth

    @property
    def headers(self):
        return {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
            "User-Agent": self.user_agent,
        }

    @property
    def expires_at(self):
        """Return the expires_at value from the token as a UTC datetime."""
        return datetime.datetime.utcfromtimestamp(self.token["expires_at"])

    def expired(self, seconds=30, now=None):
        """Check if the token has expired yet.
        :param seconds: the minimum number of seconds allowed before expiry.
        """
        if now is None:
            now = datetime.datetime.utcnow()
        # Allow a bit of time for clock differences and round trip times
        # to prevent false negatives. If users want the precise expiry,
        # they can use self.expires_at.
        return (self.expires_at - now) < datetime.timedelta(seconds=seconds)

    def refresh(self):
        """Obtain a refreshed token. Note that `offline_access` must be
        included in scope in order for a token to be refreshable.
        """
        if not self.token:
            raise XeroException(None, "Cannot refresh token, no token is present.")
        elif not self.client_secret:
            raise XeroException(
                None, "Cannot refresh token, " "client_secret must be supplied."
            )
        elif not self.token.get("refresh_token"):
            raise XeroException(
                None,
                "Token cannot be refreshed, was " "`offline_access` included in scope?",
            )
        session = OAuth2Session(
            client_id=self.client_id, scope=self.scope, token=self.token
        )
        auth = requests.auth.HTTPBasicAuth(self.client_id, self.client_secret)
        token = session.refresh_token(
            XERO_OAUTH2_TOKEN_URL, auth=auth, headers=self.headers
        )
        self._init_oauth(token)
        return token

    def get_tenants(self):
        """
        Get the list of tenants (Xero Organisations) to which this token grants access.
        """
        connection_url = self.base_url + XERO_OAUTH2_CONNECTIONS_URL

        response = requests.get(connection_url, auth=self.oauth, headers=self.headers)
        if response.status_code == 200:
            return response.json()
        else:
            self._handle_error_response(response)

    def set_default_tenant(self):
        """A quick way to set the tenant to the first in the list of available
        connections.
        """
        try:
            self.tenant_id = self.get_tenants()[0]["tenantId"]
        except IndexError:
            raise XeroException(
                None,
                "This app is not authorised to access any Xero Organisations. Did the "
                "scopes requested include access to organisation data, or has access "
                "to the organisation(s) been removed?",
            )

    @staticmethod
    def _handle_error_response(response):
        if response.status_code == 400:
            raise XeroBadRequest(response)

        elif response.status_code == 401:
            raise XeroUnauthorized(response)

        elif response.status_code == 403:
            raise XeroForbidden(response)

        elif response.status_code == 404:
            raise XeroNotFound(response)

        elif response.status_code == 500:
            raise XeroInternalError(response)

        elif response.status_code == 501:
            raise XeroNotImplemented(response)

        elif response.status_code == 503:
            # Two 503 responses are possible. Rate limit errors
            # return encoded content; offline errors don't.
            # If you parse the response text and there's nothing
            # encoded, it must be a not-available error.
            payload = parse_qs(response.text)
            if payload:
                raise XeroRateLimitExceeded(response, payload)
            else:
                raise XeroNotAvailable(response)
        else:
            raise XeroExceptionUnknown(response)



from __future__ import unicode_literals

import datetime
import re
import requests
import six

DATE = re.compile(
    r"^(\/Date\((?P<timestamp>-?\d+)((?P<offset_h>[-+]\d\d)(?P<offset_m>\d\d))?\)\/)"
    r"|"
    r"((?P<year>\d{4})-(?P<month>[0-2]\d)-0?(?P<day>[0-3]\d)"
    r"T"
    r"(?P<hour>[0-5]\d):(?P<minute>[0-5]\d):(?P<second>[0-6]\d))$"
)

OBJECT_NAMES = {
    "Addresses": "Address",
    "Attachments": "Attachment",
    "Accounts": "Account",
    "BankAccounts": "BankAccount",
    "BankTransactions": "BankTransaction",
    "BankTransfers": "BankTransfer",
    "BrandingThemes": "BrandingTheme",
    "BatchPayments": "BatchPayment",
    "ContactGroups": "ContactGroup",
    "ContactPersons": "ContactPerson",
    "Contacts": "Contact",
    "CreditNotes": "CreditNote",
    "Currencies": "Currency",
    "DeductionLines": "DeductionLine",
    "Employees": "Employee",
    "EarningsLines": "EarningsLine",
    "ExpenseClaims": "ExpenseClaim",
    "Invoices": "Invoice",
    "Items": "Item",
    "Journals": "Journal",
    "LeaveAccrualLines": "LeaveAccrualLine",
    "LeaveBalances": "LeaveBalance",
    "LeaveLines": "LeaveLine",
    "ManualJournals": "ManualJournal",
    "Organisation": "Organisation",
    "Overpayments": "Overpayment",
    "Payments": "Payment",
    "PayrollCalendars": "PayrollCalendar",
    "PayRuns": "PayRun",
    "Phones": "Phone",
    "Prepayments": "Prepayment",
    "Projects": "Project",
    "Receipts": "Receipt",
    "ReimbursementLines": "ReimbursementLine",
    "RepeatingInvoices": "RepeatingInvoice",
    "Reports": "Report",
    "SuperannuationLines": "SuperannuationLine",
    "SuperLines": "SuperLine",
    "SuperMemberships": "SuperMembership",
    "TaxComponents": "TaxComponent",
    "TaxLines": "TaxLine",
    "TaxRates": "TaxRate",
    "TimesheetEarningsLines": "TimesheetEarningsLine",
    "TrackingCategories": "TrackingCategory",
    "Tracking": "TrackingCategory",
    "Users": "User",
    "Associations": "Association",
    "Files": "File",
    "Folders": "Folder",
    "Inbox": "Inbox",
    "LineItems": "LineItem",
    "JournalLines": "JournalLine",
    "PurchaseOrders": "PurchaseOrder",
    "Quotes": "Quote",
}


def isplural(word):
    return word in OBJECT_NAMES.keys()


def singular(word):
    return OBJECT_NAMES.get(word)


def parse_date(string, force_datetime=False):
    """ Takes a Xero formatted date, e.g. /Date(1426849200000+1300)/"""
    matches = DATE.match(string)
    if not matches:
        return None

    values = dict(
        [
            (k, v if v[0] in "+-" else int(v))
            for k, v in matches.groupdict().items()
            if v and int(v)
        ]
    )

    if "timestamp" in values:
        value = datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(
            hours=int(values.get("offset_h", 0)),
            minutes=int(values.get("offset_m", 0)),
            seconds=int(values["timestamp"]) / 1000.0,
        )
        return value

    # I've made an assumption here, that a DateTime value will not
    # ever be YYYY-MM-DDT00:00:00, which is probably bad. I'm not
    # really sure how to handle this, other than to hard-code the
    # names of the field that are actually Date rather than DateTime.
    if len(values) > 3 or force_datetime:
        return datetime.datetime(**values)

    # Sometimes Xero returns Date(0+0000), so we end up with no
    # values. Return None for this case
    if not values:
        return None

    return datetime.date(**values)


def json_load_object_hook(dct):
    """ Hook for json.parse(...) to parse Xero date formats.
    """
    for key, value in dct.items():
        if isinstance(value, six.string_types):
            value = parse_date(value)
            if value:
                dct[key] = value

    return dct


def resolve_user_agent(user_agent, default_override=None):
    from xero import __version__ as VERSION

    return (
        user_agent
        or default_override
        or "pyxero/%s " % VERSION + requests.utils.default_user_agent()
    )



from __future__ import unicode_literals

import unittest
from datetime import date
from mock import Mock, patch

from xero import Xero
from xero.exceptions import (
    XeroBadRequest,
    XeroExceptionUnknown,
    XeroForbidden,
    XeroInternalError,
    XeroNotAvailable,
    XeroNotFound,
    XeroNotImplemented,
    XeroRateLimitExceeded,
    XeroUnauthorized,
)

from . import mock_data


class ExceptionsTest(unittest.TestCase):
    @patch("requests.put")
    def test_bad_request(self, r_put):
        "Data with validation errors raises a bad request exception"
        # Verified response from the live API
        head = dict()
        head["content-type"] = "text/xml; charset=utf-8"
        r_put.return_value = Mock(
            status_code=400,
            encoding="utf-8",
            text=mock_data.bad_request_text,
            headers=head,
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.invoices.put(
                {
                    "Type": "ACCREC",
                    "LineAmountTypes": "Exclusive",
                    "Date": date(2013, 4, 29),
                    "DueDate": date(2013, 4, 29),
                    "Reference": "Order # 123456",
                    "Status": "PAID",
                    "AmountPaid": "19.05",
                    "TotalTax": "1.05",
                    "AmountDue": "0.00",
                    "Total": "19.05",
                    "SubTotal": "18.00",
                }
            )
            self.fail("Should raise a XeroBadRequest.")

        except XeroBadRequest as e:
            # Error messages have been extracted
            self.assertEqual(str(e), "A validation exception occurred")
            self.assertEqual(
                e.errors,
                [
                    "One or more line items must be specified",
                    "Invoice not of valid status for creation",
                    "A Contact must be specified for this type of transaction",
                ],
            )

            # The response has also been stored
            self.assertEqual(e.response.status_code, 400)
            self.assertTrue(e.response.text.startswith("<ApiException"))
        except Exception as e:
            self.fail("Should raise a XeroBadRequest, not %s" % e)

    @patch("requests.put")
    def test_bad_request_invalid_response(self, r_put):
        "If the error response from the backend is malformed (or truncated), raise a XeroExceptionUnknown"
        head = {"content-type": "text/xml; charset=utf-8"}

        # Same error as before, but the response got cut off prematurely
        bad_response = mock_data.bad_request_text[:1000]

        r_put.return_value = Mock(
            status_code=400, encoding="utf-8", text=bad_response, headers=head
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        with self.assertRaises(
            XeroExceptionUnknown, msg="Should raise a XeroExceptionUnknown"
        ):
            xero.invoices.put(
                {
                    "Type": "ACCREC",
                    "LineAmountTypes": "Exclusive",
                    "Date": date(2013, 4, 29),
                    "DueDate": date(2013, 4, 29),
                    "Reference": "Order # 123456",
                    "Status": "PAID",
                    "AmountPaid": "19.05",
                    "TotalTax": "1.05",
                    "AmountDue": "0.00",
                    "Total": "19.05",
                    "SubTotal": "18.00",
                }
            )

    @patch("requests.get")
    def test_unregistered_app(self, r_get):
        "An app without a signature raises a BadRequest exception, but with HTML payload"
        # Verified response from the live API
        head = dict()
        head["content-type"] = "text/html; charset=utf-8"
        r_get.return_value = Mock(
            status_code=400,
            text="oauth_problem=signature_method_rejected&oauth_problem_advice=No%20certificates%20have%20been%20registered%20for%20the%20consumer",
            headers=head,
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.all()
            self.fail("Should raise a XeroUnauthorized.")

        except XeroBadRequest as e:
            # Error messages have been extracted
            self.assertEqual(
                str(e), "No certificates have been registered for the consumer"
            )
            self.assertEqual(e.errors[0], "signature_method_rejected")

            # The response has also been stored
            self.assertEqual(e.response.status_code, 400)
            self.assertEqual(
                e.response.text,
                "oauth_problem=signature_method_rejected&oauth_problem_advice=No%20certificates%20have%20been%20registered%20for%20the%20consumer",
            )

        except Exception as e:
            self.fail("Should raise a XeroBadRequest, not %s" % e)

    @patch("requests.get")
    def test_unauthorized_invalid(self, r_get):
        "A session with an invalid token raises an unauthorized exception"
        # Verified response from the live API
        r_get.return_value = Mock(
            status_code=401,
            text="oauth_problem=signature_invalid&oauth_problem_advice=Failed%20to%20validate%20signature",
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.all()
            self.fail("Should raise a XeroUnauthorized.")

        except XeroUnauthorized as e:
            # Error messages have been extracted
            self.assertEqual(str(e), "Failed to validate signature")
            self.assertEqual(e.errors[0], "signature_invalid")

            # The response has also been stored
            self.assertEqual(e.response.status_code, 401)
            self.assertEqual(
                e.response.text,
                "oauth_problem=signature_invalid&oauth_problem_advice=Failed%20to%20validate%20signature",
            )
        except Exception as e:
            self.fail("Should raise a XeroUnauthorized, not %s" % e)

    @patch("requests.get")
    def test_unauthorized_expired(self, r_get):
        "A session with an expired token raises an unauthorized exception"
        # Verified response from the live API
        r_get.return_value = Mock(
            status_code=401,
            text="oauth_problem=token_expired&oauth_problem_advice=The%20access%20token%20has%20expired",
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.all()
            self.fail("Should raise a XeroUnauthorized.")

        except XeroUnauthorized as e:
            # Error messages have been extracted
            self.assertEqual(str(e), "The access token has expired")
            self.assertEqual(e.errors[0], "token_expired")

            # The response has also been stored
            self.assertEqual(e.response.status_code, 401)
            self.assertEqual(
                e.response.text,
                "oauth_problem=token_expired&oauth_problem_advice=The%20access%20token%20has%20expired",
            )
        except Exception as e:
            self.fail("Should raise a XeroUnauthorized, not %s" % e)

    @patch("requests.get")
    def test_forbidden(self, r_get):
        "In case of an SSL failure, a Forbidden exception is raised"
        # This is unconfirmed; haven't been able to verify this response from API.
        r_get.return_value = Mock(
            status_code=403, text="The client SSL certificate was not valid."
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.all()
            self.fail("Should raise a XeroForbidden.")

        except XeroForbidden as e:
            # Error messages have been extracted
            self.assertEqual(str(e), "The client SSL certificate was not valid.")

            # The response has also been stored
            self.assertEqual(e.response.status_code, 403)
            self.assertEqual(
                e.response.text, "The client SSL certificate was not valid."
            )
        except Exception as e:
            self.fail("Should raise a XeroForbidden, not %s" % e)

    @patch("requests.get")
    def test_not_found(self, r_get):
        "If you request an object that doesn't exist, a Not Found exception is raised"
        # Verified response from the live API
        r_get.return_value = Mock(
            status_code=404, text="The resource you're looking for cannot be found"
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.get(id="deadbeef")
            self.fail("Should raise a XeroNotFound.")

        except XeroNotFound as e:
            # Error messages have been extracted
            self.assertEqual(str(e), "The resource you're looking for cannot be found")

            # The response has also been stored
            self.assertEqual(e.response.status_code, 404)
            self.assertEqual(
                e.response.text, "The resource you're looking for cannot be found"
            )
        except Exception as e:
            self.fail("Should raise a XeroNotFound, not %s" % e)

    @patch("requests.get")
    def test_internal_error(self, r_get):
        "In case of an SSL failure, a Forbidden exception is raised"
        # This is unconfirmed; haven't been able to verify this response from API.
        r_get.return_value = Mock(
            status_code=500,
            text="An unhandled error with the Xero API occurred. Contact the Xero API team if problems persist.",
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.all()
            self.fail("Should raise a XeroInternalError.")

        except XeroInternalError as e:
            # Error messages have been extracted
            self.assertEqual(
                str(e),
                "An unhandled error with the Xero API occurred. Contact the Xero API team if problems persist.",
            )

            # The response has also been stored
            self.assertEqual(e.response.status_code, 500)
            self.assertEqual(
                e.response.text,
                "An unhandled error with the Xero API occurred. Contact the Xero API team if problems persist.",
            )
        except Exception as e:
            self.fail("Should raise a XeroInternalError, not %s" % e)

    @patch("requests.post")
    def test_not_implemented(self, r_post):
        "In case of an SSL failure, a Forbidden exception is raised"
        # Verified response from the live API
        r_post.return_value = Mock(
            status_code=501, encoding="utf-8", text=mock_data.not_implemented_text
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.organisations.save({})
            self.fail("Should raise a XeroNotImplemented.")

        except XeroNotImplemented as e:
            # Error messages have been extracted
            self.assertEqual(str(e), "The Api Method called is not implemented")

            # The response has also been stored
            self.assertEqual(e.response.status_code, 501)
            self.assertTrue(e.response.text.startswith, "<ApiException")
        except Exception as e:
            self.fail("Should raise a XeroNotImplemented, not %s" % e)

    @patch("requests.get")
    def test_rate_limit_exceeded(self, r_get):
        "If you exceed the rate limit, an exception is raised."
        # Response based off Xero documentation; not confirmed by reality.
        r_get.return_value = Mock(
            status_code=503,
            text="oauth_problem=rate%20limit%20exceeded&oauth_problem_advice=please%20wait%20before%20retrying%20the%20xero%20api",
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.all()
            self.fail("Should raise a XeroRateLimitExceeded.")

        except XeroRateLimitExceeded as e:
            # Error messages have been extracted
            self.assertEqual(str(e), "please wait before retrying the xero api")
            self.assertEqual(e.errors[0], "rate limit exceeded")

            # The response has also been stored
            self.assertEqual(e.response.status_code, 503)
            self.assertEqual(
                e.response.text,
                "oauth_problem=rate%20limit%20exceeded&oauth_problem_advice=please%20wait%20before%20retrying%20the%20xero%20api",
            )
        except Exception as e:
            self.fail("Should raise a XeroRateLimitExceeded, not %s" % e)

    @patch("requests.get")
    def test_not_available(self, r_get):
        "If Xero goes down for maintenance, an exception is raised"
        # Response based off Xero documentation; not confirmed by reality.
        r_get.return_value = Mock(
            status_code=503, text="The Xero API is currently offline for maintenance"
        )

        credentials = Mock(base_url="")
        xero = Xero(credentials)

        try:
            xero.contacts.all()
            self.fail("Should raise a XeroNotAvailable.")

        except XeroNotAvailable as e:
            # Error messages have been extracted
            self.assertEqual(
                str(e), "The Xero API is currently offline for maintenance"
            )

            # The response has also been stored
            self.assertEqual(e.response.status_code, 503)
            self.assertEqual(
                e.response.text, "The Xero API is currently offline for maintenance"
            )
        except Exception as e:
            self.fail("Should raise a XeroNotAvailable, not %s" % e)



from __future__ import unicode_literals

# exceptions.py

bad_request_text = """<ApiException xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <ErrorNumber>10</ErrorNumber>
  <Type>ValidationException</Type>
  <Message>A validation exception occurred</Message>
  <Elements>
    <DataContractBase xsi:type="Invoice">
      <ValidationErrors>
        <ValidationError>
          <Message>One or more line items must be specified</Message>
        </ValidationError>
        <ValidationError>
          <Message>Invoice not of valid status for creation</Message>
        </ValidationError>
        <ValidationError>
          <Message>A Contact must be specified for this type of transaction</Message>
        </ValidationError>
      </ValidationErrors>
      <Warnings />
      <Date>2013-04-29T00:00:00</Date>
      <DueDate>2013-04-29T00:00:00</DueDate>
      <BrandingThemeID xsi:nil="true" />
      <Status>PAID</Status>
      <LineAmountTypes>Exclusive</LineAmountTypes>
      <LineItems />
      <SubTotal>18.00</SubTotal>
      <TotalTax>1.05</TotalTax>
      <Total>19.05</Total>
      <UpdatedDateUTC xsi:nil="true" />
      <CurrencyCode>AUD</CurrencyCode>
      <FullyPaidOnDate xsi:nil="true" />
      <Type>ACCREC</Type>
      <InvoiceID>00000000-0000-0000-0000-000000000000</InvoiceID>
      <Reference>Order # 123456</Reference>
      <Payments />
      <CreditNotes />
      <AmountDue>0.00</AmountDue>
      <AmountPaid>19.05</AmountPaid>
      <AmountCredited xsi:nil="true" />
      <SentToContact xsi:nil="true" />
      <CurrencyRate xsi:nil="true" />
      <TotalDiscount xsi:nil="true" />
      <HasAttachments xsi:nil="true" />
      <Attachments />
    </DataContractBase>
  </Elements>
</ApiException>"""

not_implemented_text = """<ApiException xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <ErrorNumber>20</ErrorNumber>
    <Type>ApiMethodNotImplementedException</Type>
    <Message>The Api Method called is not implemented</Message>
</ApiException>"""


# manager.py

unicode_content_text = """<Response xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <Id>dbb54b2b-8fdb-4277-ad03-2df50ce760fa</Id>
  <Status>OK</Status>
  <ProviderName>TradesCloud</ProviderName>
  <DateTimeUTC>2013-05-31T06:07:35.3732465Z</DateTimeUTC>
  <Contacts>
    <Contact>
      <ContactID>755f1475-d255-43a8-bedc-5ea7fd26c71f</ContactID>
      <ContactStatus>ACTIVE</ContactStatus>
      <Name>Yarra Transport</Name>
      <FirstName>John</FirstName>
      <LastName>S\xfcrname</LastName>
      <EmailAddress>rayong@yarratransport.co</EmailAddress>
      <Addresses>
        <Address>
          <AddressType>STREET</AddressType>
        </Address>
        <Address>
          <AddressType>POBOX</AddressType>
          <AddressLine1>P O Box 5678</AddressLine1>
          <City>Melbourne</City>
          <PostalCode>3133</PostalCode>
        </Address>
      </Addresses>
      <Phones>
        <Phone>
          <PhoneType>DDI</PhoneType>
        </Phone>
        <Phone>
          <PhoneType>DEFAULT</PhoneType>
          <PhoneNumber>12344321</PhoneNumber>
          <PhoneAreaCode>03</PhoneAreaCode>
        </Phone>
        <Phone>
          <PhoneType>FAX</PhoneType>
        </Phone>
        <Phone>
          <PhoneType>MOBILE</PhoneType>
        </Phone>
      </Phones>
      <UpdatedDateUTC>2013-05-31T06:04:20.78</UpdatedDateUTC>
      <ContactGroups>
        <ContactGroup>
          <ContactGroupID>26fcca8d-a03b-4968-a80a-a463d5bf30ee</ContactGroupID>
          <Name>Support Clients (monthly)</Name>
          <Status>ACTIVE</Status>
        </ContactGroup>
      </ContactGroups>
      <IsSupplier>false</IsSupplier>
      <IsCustomer>true</IsCustomer>
    </Contact>
  </Contacts>
</Response>
"""



from __future__ import unicode_literals

import requests

from .basemanager import BaseManager
from .constants import XERO_PAYROLL_URL
from .utils import singular


class PayrollManager(BaseManager):
    def __init__(self, name, credentials, unit_price_4dps=False, user_agent=None):
        from xero import __version__ as VERSION

        self.credentials = credentials
        self.name = name
        self.base_url = credentials.base_url + XERO_PAYROLL_URL
        self.extra_params = {"unitdp": 4} if unit_price_4dps else {}
        self.singular = singular(name)

        if user_agent is None:
            self.user_agent = (
                "pyxero/%s " % VERSION + requests.utils.default_user_agent()
            )
        else:
            self.user_agent = user_agent

        for method_name in self.DECORATED_METHODS:
            method = getattr(self, "_%s" % method_name)
            setattr(self, method_name, self._get_data(method))



XERO_BASE_URL = "https://api.xero.com"

REQUEST_TOKEN_URL = "/oauth/RequestToken"
AUTHORIZE_URL = "/oauth/Authorize"
ACCESS_TOKEN_URL = "/oauth/AccessToken"
XERO_API_URL = "/api.xro/2.0"
XERO_FILES_URL = "/files.xro/1.0"
XERO_PAYROLL_URL = "/payroll.xro/1.0"
XERO_PROJECTS_URL = "/projects.xro/2.0"

XERO_OAUTH2_AUTHORIZE_URL = "https://login.xero.com/identity/connect/authorize"
XERO_OAUTH2_TOKEN_URL = "https://identity.xero.com/connect/token"
XERO_OAUTH2_CONNECTIONS_URL = "/connections"


class XeroScopes:
    # Offline Access
    OFFLINE_ACCESS = "offline_access"

    # OpenID connection
    OPENID = "openid"
    PROFILE = "profile"
    EMAIL = "email"

    # Accounting API
    ACCOUNTING_TRANSACTIONS = "accounting.transactions"
    ACCOUNTING_TRANSACTIONS_READ = "accounting.transactions.read"
    ACCOUNTING_REPORTS_READ = "accounting.reports.read"
    ACCOUNTING_JOURNALS_READ = "accounting.journals.read"
    ACCOUNTING_SETTINGS = "accounting.settings"
    ACCOUNTING_SETTINGS_READ = "accounting.settings.read"
    ACCOUNTING_CONTACTS = "accounting.contacts"
    ACCOUNTING_CONTACTS_READ = "accounting.contacts.read"
    ACCOUNTING_ATTACHMENTS = "accounting.attachments"
    ACCOUNTING_ATTACHMENTS_READ = "accounting.attachments.read"

    # Payroll API
    PAYROLL_EMPLOYEES = "payroll.employees"
    PAYROLL_EMPLOYEES_READ = "payroll.employees.read"
    PAYROLL_PAYRUNS = "payroll.payruns"
    PAYROLL_PAYRUNS_READ = "payroll.payruns.read"
    PAYROLL_PAYSLIP = "payroll.payslip"
    PAYROLL_PAYSLIP_READ = "payroll.payslip.read"
    PAYROLL_TIMESHEETS = "payroll.timesheets"
    PAYROLL_TIMESHEETS_READ = "payroll.timesheets.read"
    PAYROLL_SETTINGS = "payroll.settings"
    PAYROLL_SETTINGS_READ = "payroll.settings.read"

    # Files API
    FILES = "files"
    FILES_READ = "files.read"

    # Asssets API
    ASSETS = "assets"
    ASSETS_READ = "assets.read"

    # Projects API
    PROJECTS = "projects"
    PROJECTS_READ = "projects.read"

    # Restricted Scopes
    PAYMENTSERVICES = "paymentservices"
    BANKFEEDS = "bankfeeds"



#!/usr/bin/env python

"""Upload receipts to Xero from a CSV file"""

# Contributed by @davidbitton, https://github.com/freakboy3742/pyxero/pull/48

# http://developer.xero.com/documentation/api/receipts/

import argparse
import csv
import os.path
import pickle
import sys
import webbrowser

from xero import Xero
from xero.auth import PublicCredentials

# from config import xero_config
xero_config = {
    "consumer_key": "XXX",
    "consumer_secret": "XXX",
}


argp = argparse.ArgumentParser(description=__doc__)
argp.add_argument('-i', '--input-file', dest='input_file', default='-', help='input file name')
argp.add_argument('-c', '--credentials-file', dest='credentials_file', default='credentials.pickle', help='credentials file name')
args = argp.parse_args()

# Connect to Xero
if os.path.isfile(args.credentials_file):
    # Use cached credentials
    with open(args.credentials_file, 'r') as credentials_fh:
        credentials_state = pickle.loads(credentials_fh.read())
        credentials = PublicCredentials(**credentials_state)
else:
    credentials = PublicCredentials(xero_config['consumer_key'], xero_config['consumer_secret'])
    print(credentials.url)
    webbrowser.open(credentials.url)
    verifier = input('Enter auth code: ')
    credentials.verify(verifier)
    with open(args.credentials_file, 'w') as credentials_fh:
        pickle.dump(credentials.state, credentials_fh)
xero = Xero(credentials)

# Get userids matching user email
userids = {}
users = xero.users.all()
# print(users)
for user in users:
    userids[user['EmailAddress']] = user['UserID']

# Get receipts and add them to Xero
if args.input_file == '-':
    in_fh = sys.stdin
else:
    in_fh = open(args.input_file, 'r')

# Date,UserEmail,ContactName,Reference,Description,UnitAmount,AccountCode
reader = csv.DictReader(in_fh, delimiter=',')
receipts = []
for i in reader:
    # print(i)

    user_email = i['UserEmail']
    if user_email not in userids:
        print("Unknown user: %s" % (user_email,))
        continue
    userid = userids[i['UserEmail']],

    # TODO: Validate input data

    data = {
        'Date': i['Date'],
        'Contact': {'Name': i['ContactName']},
        'Reference': i['Reference'],
        'LineItems': [
            {'Description': i['Description'], 'UnitAmount': i['UnitAmount'], 'AccountCode': i['AccountCode']},
        ],
        'User': {'UserID': userids[i['UserEmail']]},
    }
    # print(data)
    # xml = xero.receipts._prepare_data_for_save(data)
    # print(xml)
    receipts.append(data)

# print(receipts)

if receipts:
    results = xero.receipts.put(receipts)
    # import pprint
    # pp = pprint.PrettyPrinter(depth=6)
    # pp.pprint(results)



#!/usr/bin/env python
import io
import re

from setuptools import setup

with io.open('./xero/__init__.py', encoding='utf8') as version_file:
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file.read(), re.M)
    if version_match:
        version = version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


with io.open('README.md', encoding='utf8') as readme:
    long_description = readme.read()


setup(
    name='pyxero',
    version=version,
    description='Python API for accessing the REST API of the Xero accounting tool.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Russell Keith-Magee',
    author_email='russell@keith-magee.com',
    url='http://github.com/freakboy3742/pyxero',
    packages=['xero', ],
    install_requires=[
        'six>=1.8.0',
        'requests>=1.1.0',
        'requests-oauthlib>=0.3.0',
        'python-dateutil>=2.8',
        'PyJWT>=1.6.4',  # This is required as part of oauthlib but doesn't seem to get included sometimes.
        'cryptography>=1.3.1',  # As above, but fixes issue with missing module imports not picked up for some reason.
    ],
    tests_require=[
        'mock',
    ],
    license='New BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Office/Business :: Financial :: Accounting',
    ],
    test_suite="tests",
)



from __future__ import unicode_literals

import json
import requests
import six
from datetime import datetime
from six.moves.urllib.parse import parse_qs
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.parsers.expat import ExpatError

from .auth import OAuth2Credentials
from .exceptions import (
    XeroBadRequest,
    XeroExceptionUnknown,
    XeroForbidden,
    XeroInternalError,
    XeroNotAvailable,
    XeroNotFound,
    XeroNotImplemented,
    XeroRateLimitExceeded,
    XeroTenantIdNotSet,
    XeroUnauthorized,
)
from .utils import isplural, json_load_object_hook, singular


class BaseManager(object):
    DECORATED_METHODS = (
        "get",
        "save",
        "filter",
        "all",
        "put",
        "delete",
        "get_history",
        "put_history",
        "get_attachments",
        "get_attachment_data",
        "put_attachment_data",
    )
    OBJECT_DECORATED_METHODS = {
        "Invoices": ["email", "online_invoice"],
    }
    DATETIME_FIELDS = (
        "UpdatedDateUTC",
        "Updated",
        "FullyPaidOnDate",
        "DateTimeUTC",
        "CreatedDateUTC",
        "JournalDate",
    )
    DATE_FIELDS = (
        "DueDate",
        "Date",
        "PaymentDate",
        "StartDate",
        "EndDate",
        "PeriodLockDate",
        "DateOfBirth",
        "OpeningBalanceDate",
        "PaymentDueDate",
        "ReportingDate",
        "DeliveryDate",
        "ExpectedArrivalDate",
    )
    BOOLEAN_FIELDS = (
        "IsSupplier",
        "IsCustomer",
        "IsDemoCompany",
        "PaysTax",
        "IsAuthorisedToApproveTimesheets",
        "IsAuthorisedToApproveLeave",
        "HasHELPDebt",
        "AustralianResidentForTaxPurposes",
        "TaxFreeThresholdClaimed",
        "HasSFSSDebt",
        "EligibleToReceiveLeaveLoading",
        "IsExemptFromTax",
        "IsExemptFromSuper",
        "SentToContact",
        "IsSubscriber",
        "HasAttachments",
        "ShowOnCashBasisReports",
        "IncludeInEmails",
        "SentToContact",
        "CanApplyToRevenue",
        "IsReconciled",
        "EnablePaymentsToAccount",
        "ShowInExpenseClaims",
        "DiscountEnteredAsPercent",
        "IsPurchased",
        "IsSold",
        "IsTrackedAsInventory",
    )
    DECIMAL_FIELDS = (
        "Hours",
        "NumberOfUnit",
    )
    INTEGER_FIELDS = (
        "FinancialYearEndDay",
        "FinancialYearEndMonth",
    )
    NO_SEND_FIELDS = (
        "UpdatedDateUTC",
        "HasValidationErrors",
        "IsDiscounted",
        "DateString",
        "HasErrors",
        "DueDateString",
    )
    OPERATOR_MAPPINGS = {
        "gt": ">",
        "lt": "<",
        "lte": "<=",
        "gte": ">=",
        "ne": "!=",
    }

    def __init__(self):
        pass

    def dict_to_xml(self, root_elm, data):
        for key in data.keys():
            # Xero will complain if we send back these fields.
            if key in self.NO_SEND_FIELDS:
                continue

            sub_data = data[key]
            elm = SubElement(root_elm, key)

            # Key references a dict. Unroll the dict
            # as it's own XML node with subnodes
            if isinstance(sub_data, dict):
                self.dict_to_xml(elm, sub_data)

            # Key references a list/tuple
            elif isinstance(sub_data, list) or isinstance(sub_data, tuple):
                # key name is a plural. This means each item
                # in the list needs to be wrapped in an XML
                # node that is a singular version of the list name.
                if isplural(key):
                    for d in sub_data:
                        self.dict_to_xml(SubElement(elm, singular(key)), d)

                # key name isn't a plural. Just insert the content
                # as an XML node with subnodes
                else:
                    for d in sub_data:
                        self.dict_to_xml(elm, d)

            # Normal element - just insert the data.
            else:
                if key in self.BOOLEAN_FIELDS:
                    val = "true" if sub_data else "false"
                elif key in self.DATE_FIELDS:
                    val = sub_data.strftime("%Y-%m-%dT%H:%M:%S")
                else:
                    val = six.text_type(sub_data)
                elm.text = val

        return root_elm

    def _prepare_data_for_save(self, data):
        if isinstance(data, list) or isinstance(data, tuple):
            root_elm = Element(self.name)
            for d in data:
                sub_elm = SubElement(root_elm, self.singular)
                self.dict_to_xml(sub_elm, d)
        else:
            root_elm = self.dict_to_xml(Element(self.singular), data)

        # In python3 this seems to return a bytestring
        return six.u(tostring(root_elm))

    def _parse_api_response(self, response, resource_name):
        data = json.loads(response.text, object_hook=json_load_object_hook)
        assert data["Status"] == "OK", (
            "Expected the API to say OK but received %s" % data["Status"]
        )

        try:
            return data[resource_name]
        except KeyError:
            return data

    def _get_data(self, func):
        """ This is the decorator for our DECORATED_METHODS.
        Each of the decorated methods must return:
            uri, params, method, body, headers, singleobject
        """

        def wrapper(*args, **kwargs):
            timeout = kwargs.pop("timeout", None)

            uri, params, method, body, headers, singleobject = func(*args, **kwargs)

            if headers is None:
                headers = {}

            if isinstance(self.credentials, OAuth2Credentials):
                if self.credentials.tenant_id:
                    headers["Xero-tenant-id"] = self.credentials.tenant_id
                else:
                    raise XeroTenantIdNotSet

            # Use the JSON API by default, but remember we might request a PDF (application/pdf)
            # so don't force the Accept header.
            if "Accept" not in headers:
                headers["Accept"] = "application/json"

            # Set a user-agent so Xero knows the traffic is coming from pyxero
            # or individual user/partner
            headers["User-Agent"] = self.user_agent

            response = getattr(requests, method)(
                uri,
                data=body,
                headers=headers,
                auth=self.credentials.oauth,
                params=params,
                timeout=timeout,
            )

            if response.status_code == 200:
                # If we haven't got XML or JSON, assume we're being returned a
                # binary file
                if not response.headers["content-type"].startswith("application/json"):
                    return response.content

                return self._parse_api_response(response, self.name)

            elif response.status_code == 204:
                return response.content

            elif response.status_code == 400:
                try:
                    raise XeroBadRequest(response)
                except (ValueError, ExpatError):
                    raise XeroExceptionUnknown(
                        response, msg="Unable to parse Xero API response"
                    )

            elif response.status_code == 401:
                raise XeroUnauthorized(response)

            elif response.status_code == 403:
                raise XeroForbidden(response)

            elif response.status_code == 404:
                raise XeroNotFound(response)

            elif response.status_code == 500:
                raise XeroInternalError(response)

            elif response.status_code == 501:
                raise XeroNotImplemented(response)

            elif response.status_code == 503:
                # Two 503 responses are possible. Rate limit errors
                # return encoded content; offline errors don't.
                # If you parse the response text and there's nothing
                # encoded, it must be a not-available error.
                payload = parse_qs(response.text)
                if payload:
                    raise XeroRateLimitExceeded(response, payload)
                else:
                    raise XeroNotAvailable(response)
            else:
                raise XeroExceptionUnknown(response)

        return wrapper

    def _get(self, id, headers=None, params=None):
        uri = "/".join([self.base_url, self.name, id])
        uri_params = self.extra_params.copy()
        uri_params.update(params if params else {})
        return uri, uri_params, "get", None, headers, True

    def _get_history(self, id):
        uri = "/".join([self.base_url, self.name, id, "history"]) + "/"
        return uri, {}, "get", None, None, False

    def _get_attachments(self, id):
        """Retrieve a list of attachments associated with this Xero object."""
        uri = "/".join([self.base_url, self.name, id, "Attachments"]) + "/"
        return uri, {}, "get", None, None, False

    def _get_attachment_data(self, id, filename):
        """
        Retrieve the contents of a specific attachment (identified by filename).
        """
        uri = "/".join([self.base_url, self.name, id, "Attachments", filename])
        return uri, {}, "get", None, None, False

    def get_attachment(self, id, filename, file):
        """
        Retrieve the contents of a specific attachment (identified by filename).

        Writes data to file object, returns length of data written.
        """
        data = self.get_attachment_data(id, filename)
        file.write(data)
        return len(data)

    def _email(self, id):
        uri = "/".join([self.base_url, self.name, id, "Email"])
        return uri, {}, "post", None, None, True

    def _online_invoice(self, id):
        uri = "/".join([self.base_url, self.name, id, "OnlineInvoice"])
        return uri, {}, "get", None, None, True

    def save_or_put(self, data, method="post", headers=None, summarize_errors=True):
        uri = "/".join([self.base_url, self.name])
        body = {"xml": self._prepare_data_for_save(data)}
        params = self.extra_params.copy()
        if not summarize_errors:
            params["summarizeErrors"] = "false"
        return uri, params, method, body, headers, False

    def _save(self, data):
        return self.save_or_put(data, method="post")

    def _put(self, data, summarize_errors=True):
        return self.save_or_put(data, method="put", summarize_errors=summarize_errors)

    def _delete(self, id):
        uri = "/".join([self.base_url, self.name, id])
        return uri, {}, "delete", None, None, False

    def _put_history_data(self, id, details):
        """Add a history note to the Xero object."""
        uri = "/".join([self.base_url, self.name, id, "history"])
        details_data = {"Details": details}
        root_elm = Element("HistoryRecord")
        self.dict_to_xml(root_elm, details_data)
        data = six.u(tostring(root_elm))
        return uri, {}, "put", data, None, False

    def _put_history(self, id, details):
        """Upload a history note to the Xero object."""
        return self._put_history_data(id, details)

    def _put_attachment_data(
        self, id, filename, data, content_type, include_online=False
    ):
        """Upload an attachment to the Xero object."""
        uri = "/".join([self.base_url, self.name, id, "Attachments", filename])
        params = {"IncludeOnline": "true"} if include_online else {}
        headers = {"Content-Type": content_type, "Content-Length": str(len(data))}
        return uri, params, "put", data, headers, False

    def put_attachment(self, id, filename, file, content_type, include_online=False):
        """Upload an attachment to the Xero object (from file object)."""
        return self.put_attachment_data(
            id, filename, file.read(), content_type, include_online=include_online
        )

    def prepare_filtering_date(self, val):
        if isinstance(val, datetime):
            val = val.strftime("%a, %d %b %Y %H:%M:%S GMT")
        else:
            val = '"%s"' % val
        return {"If-Modified-Since": val}

    def _filter(self, **kwargs):
        params = self.extra_params.copy()
        headers = None
        uri = "/".join([self.base_url, self.name])

        if kwargs:
            if "since" in kwargs:
                val = kwargs["since"]
                headers = self.prepare_filtering_date(val)
                del kwargs["since"]

            def get_filter_params(key, value):
                last_key = key.split("_")[-1]
                if last_key.endswith("ID"):
                    return 'Guid("%s")' % six.text_type(value)
                if key in self.BOOLEAN_FIELDS:
                    return "true" if value else "false"
                elif key in self.DATE_FIELDS:
                    return "DateTime(%s,%s,%s)" % (value.year, value.month, value.day)
                elif key in self.DATETIME_FIELDS:
                    return value.isoformat()
                else:
                    return '"%s"' % six.text_type(value)

            def generate_param(key, value):
                parts = key.split("__")
                field = key.replace("_", ".")
                fmt = "%s==%s"
                if len(parts) == 2:
                    # support filters:
                    # Name__Contains=John becomes Name.Contains("John")
                    if parts[1] in ["contains", "startswith", "endswith"]:
                        field = parts[0]
                        fmt = "".join(["%s.", parts[1], "(%s)"])
                    elif parts[1] in ["tolower", "toupper"]:
                        field = parts[0]
                        fmt = "".join(["%s.", parts[1], "()==%s"])
                    elif parts[1] in self.OPERATOR_MAPPINGS:
                        field = parts[0]
                        key = field
                        fmt = "%s" + self.OPERATOR_MAPPINGS[parts[1]] + "%s"
                    elif parts[1] in ["isnull"]:
                        sign = "=" if value else "!"
                        return "%s%s=null" % (parts[0], sign)
                    field = field.replace("_", ".")
                return fmt % (field, get_filter_params(key, value))

            # Move any known parameter names to the query string
            KNOWN_PARAMETERS = ["order", "offset", "page", "includeArchived"]
            for param in KNOWN_PARAMETERS:
                if param in kwargs:
                    params[param] = kwargs.pop(param)

            filter_params = []

            if "raw" in kwargs:
                raw = kwargs.pop("raw")
                filter_params.append(raw)

            # Treat any remaining arguments as filter predicates
            # Xero will break if you search without a check for null in the first position:
            # http://developer.xero.com/documentation/getting-started/http-requests-and-responses/#title3
            sortedkwargs = sorted(
                six.iteritems(kwargs), key=lambda item: -1 if "isnull" in item[0] else 0
            )
            for key, value in sortedkwargs:
                filter_params.append(generate_param(key, value))

            if filter_params:
                params["where"] = "&&".join(filter_params)

        return uri, params, "get", None, headers, False

    def _all(self):
        uri = "/".join([self.base_url, self.name])
        return uri, {}, "get", None, None, False



import os
import socketserver
from http.server import SimpleHTTPRequestHandler
from io import BytesIO
from urllib.parse import parse_qsl, urlparse

from xero import Xero
from xero.auth import PublicCredentials
from xero.exceptions import XeroException

PORT = 8000


# You should use redis or a file based persistent
# storage handler if you are running multiple servers.
OAUTH_PERSISTENT_SERVER_STORAGE = {}


class PublicCredentialsHandler(SimpleHTTPRequestHandler):
    def page_response(self, title='', body=''):
        """
        Helper to render an html page with dynamic content
        """
        f = BytesIO()
        f.write('<!DOCTYPE html">\n'.encode())
        f.write('<html>\n'.encode())
        f.write('<head><title>{}</title><head>\n'.format(title).encode())
        f.write('<body>\n<h2>{}</h2>\n'.format(title).encode())
        f.write('<div class="content">{}</div>\n'.format(body).encode())
        f.write('</body>\n</html>\n'.encode())
        length = f.tell()
        f.seek(0)
        self.send_response(200)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(length))
        self.end_headers()
        self.copyfile(f, self.wfile)
        f.close()

    def redirect_response(self, url, permanent=False):
        """
        Generate redirect response
        """
        if permanent:
            self.send_response(301)
        else:
            self.send_response(302)
        self.send_header("Location", url)
        self.end_headers()

    def do_GET(self):
        """
        Handle GET request
        """
        consumer_key = os.environ.get('XERO_CONSUMER_KEY')
        consumer_secret = os.environ.get('XERO_CONSUMER_SECRET')

        if consumer_key is None or consumer_secret is None:
            raise KeyError(
                'Please define both XERO_CONSUMER_KEY and XERO_CONSUMER_SECRET environment variables')

        print("Serving path: {}".format(self.path))
        path = urlparse(self.path)

        if path.path == '/do-auth':
            credentials = PublicCredentials(
                consumer_key, consumer_secret, callback_uri='http://localhost:8000/oauth')

            # Save generated credentials details to persistent storage
            for key, value in credentials.state.items():
                OAUTH_PERSISTENT_SERVER_STORAGE.update({key: value})

            # Redirect to Xero at url provided by credentials generation
            self.redirect_response(credentials.url)
            return

        elif path.path == '/oauth':
            params = dict(parse_qsl(path.query))
            if 'oauth_token' not in params or 'oauth_verifier' not in params or 'org' not in params:
                self.send_error(500, message='Missing parameters required.')
                return

            stored_values = OAUTH_PERSISTENT_SERVER_STORAGE
            credentials = PublicCredentials(**stored_values)

            try:
                credentials.verify(params['oauth_verifier'])

                # Resave our verified credentials
                for key, value in credentials.state.items():
                    OAUTH_PERSISTENT_SERVER_STORAGE.update({key: value})

            except XeroException as e:
                self.send_error(500, message='{}: {}'.format(e.__class__, e.message))
                return

            # Once verified, api can be invoked with xero = Xero(credentials)
            self.redirect_response('/verified')
            return

        elif path.path == '/verified':
            stored_values = OAUTH_PERSISTENT_SERVER_STORAGE
            credentials = PublicCredentials(**stored_values)

            try:
                xero = Xero(credentials)

            except XeroException as e:
                self.send_error(500, message='{}: {}'.format(e.__class__, e.message))
                return

            page_body = 'Your contacts:<br><br>'

            contacts = xero.contacts.all()

            if contacts:
                page_body += '<br>'.join([str(contact) for contact in contacts])
            else:
                page_body += 'No contacts'
            self.page_response(title='Xero Contacts', body=page_body)
            return

        SimpleHTTPRequestHandler.do_GET(self)


if __name__ == '__main__':
    httpd = socketserver.TCPServer(("", PORT), PublicCredentialsHandler)

    print("serving at port", PORT)
    httpd.serve_forever()



from __future__ import unicode_literals

from .basemanager import BaseManager
from .constants import XERO_API_URL
from .utils import resolve_user_agent, singular


class Manager(BaseManager):
    def __init__(self, name, credentials, unit_price_4dps=False, user_agent=None):
        from xero import __version__ as VERSION  # noqa

        self.credentials = credentials
        self.name = name
        self.base_url = credentials.base_url + XERO_API_URL
        self.extra_params = {"unitdp": 4} if unit_price_4dps else {}
        self.singular = singular(name)
        self.user_agent = resolve_user_agent(
            user_agent, getattr(credentials, "user_agent", None)
        )

        for method_name in self.DECORATED_METHODS:
            method = getattr(self, "_%s" % method_name)
            setattr(self, method_name, self._get_data(method))

        if self.name in self.OBJECT_DECORATED_METHODS.keys():
            object_decorated_methods = self.OBJECT_DECORATED_METHODS[self.name]
            for method_name in object_decorated_methods:
                method = getattr(self, "_%s" % method_name)
                setattr(self, method_name, self._get_data(method))



import json
import time
import unittest
from datetime import datetime, timedelta
from mock import Mock, patch
from six.moves.urllib.parse import parse_qs, urlparse

from xero.api import Xero
from xero.auth import (
    OAuth2Credentials,
    PartnerCredentials,
    PrivateCredentials,
    PublicCredentials,
)
from xero.constants import XERO_OAUTH2_AUTHORIZE_URL
from xero.exceptions import (
    XeroAccessDenied,
    XeroException,
    XeroNotVerified,
    XeroTenantIdNotSet,
    XeroUnauthorized,
)


class PublicCredentialsTest(unittest.TestCase):
    @patch("requests.post")
    def test_initial_constructor(self, r_post):
        "Initial construction causes a requst to get a request token"
        r_post.return_value = Mock(
            status_code=200, text="oauth_token=token&oauth_token_secret=token_secret"
        )

        credentials = PublicCredentials(
            consumer_key="key", consumer_secret="secret", scope="payroll.endpoint"
        )

        # A HTTP request was made
        self.assertTrue(r_post.called)

        state = credentials.state

        # Expiry times should be calculated
        self.assertIsNotNone(state.pop("oauth_authorization_expires_at"))
        self.assertIsNotNone(state.pop("oauth_expires_at"))

        self.assertEqual(
            state,
            {
                "consumer_key": "key",
                "consumer_secret": "secret",
                "oauth_token": "token",
                "oauth_token_secret": "token_secret",
                "verified": False,
                "scope": "payroll.endpoint",
            },
        )

    @patch("requests.post")
    def test_bad_credentials(self, r_post):
        "Initial construction with bad credentials raises an exception"
        r_post.return_value = Mock(
            status_code=401,
            text="oauth_problem=consumer_key_unknown&oauth_problem_advice=Consumer%20key%20was%20not%20recognised",
        )

        with self.assertRaises(XeroUnauthorized):
            PublicCredentials(consumer_key="unknown", consumer_secret="unknown")

    @patch("requests.post")
    def test_unvalidated_constructor(self, r_post):
        "Credentials with an unverified request token can be constructed"
        credentials = PublicCredentials(
            consumer_key="key",
            consumer_secret="secret",
            oauth_token="token",
            oauth_token_secret="token_secret",
        )

        self.assertEqual(
            credentials.state,
            {
                "consumer_key": "key",
                "consumer_secret": "secret",
                "oauth_token": "token",
                "oauth_token_secret": "token_secret",
                "verified": False,
            },
        )

        # No HTTP requests were made
        self.assertFalse(r_post.called)

    @patch("requests.post")
    def test_validated_constructor(self, r_post):
        "A validated set of credentials can be reconstructed"
        credentials = PublicCredentials(
            consumer_key="key",
            consumer_secret="secret",
            oauth_token="validated_token",
            oauth_token_secret="validated_token_secret",
            verified=True,
        )

        self.assertEqual(
            credentials.state,
            {
                "consumer_key": "key",
                "consumer_secret": "secret",
                "oauth_token": "validated_token",
                "oauth_token_secret": "validated_token_secret",
                "verified": True,
            },
        )

        try:
            credentials.oauth
        except XeroNotVerified:
            self.fail("Credentials should have been verified")

        # No HTTP requests were made
        self.assertFalse(r_post.called)

    @patch("requests.post")
    def test_url(self, r_post):
        "The request token URL can be obtained"
        r_post.return_value = Mock(
            status_code=200, text="oauth_token=token&oauth_token_secret=token_secret"
        )

        credentials = PublicCredentials(consumer_key="key", consumer_secret="secret")

        self.assertEqual(
            credentials.url, "https://api.xero.com/oauth/Authorize?oauth_token=token"
        )

    @patch("requests.post")
    def test_url_with_scope(self, r_post):
        "The request token URL includes the scope parameter"
        r_post.return_value = Mock(
            status_code=200, text="oauth_token=token&oauth_token_secret=token_secret"
        )

        credentials = PublicCredentials(
            consumer_key="key", consumer_secret="secret", scope="payroll.endpoint"
        )

        self.assertIn("scope=payroll.endpoint", credentials.url)

    @patch("requests.post")
    def test_configurable_url(self, r_post):
        "Test configurable API url"
        r_post.return_value = Mock(
            status_code=200, text="oauth_token=token&oauth_token_secret=token_secret"
        )

        url = "https//api-tls.xero.com"

        credentials = PublicCredentials(
            consumer_key="key", consumer_secret="secret", api_url=url
        )

        self.assertEqual(
            credentials.url, "{url}/oauth/Authorize?oauth_token=token".format(url=url)
        )

    @patch("requests.post")
    def test_verify(self, r_post):
        "Unverfied credentials can be verified"
        r_post.return_value = Mock(
            status_code=200,
            text="oauth_token=verified_token&oauth_token_secret=verified_token_secret",
        )

        credentials = PublicCredentials(
            consumer_key="key",
            consumer_secret="secret",
            oauth_token="token",
            oauth_token_secret="token_secret",
        )

        credentials.verify("verifier")

        # A HTTP request was made
        self.assertTrue(r_post.called)

        state = credentials.state

        # Expiry times should be calculated
        self.assertIsNotNone(state.pop("oauth_authorization_expires_at"))
        self.assertIsNotNone(state.pop("oauth_expires_at"))

        self.assertEqual(
            state,
            {
                "consumer_key": "key",
                "consumer_secret": "secret",
                "oauth_token": "verified_token",
                "oauth_token_secret": "verified_token_secret",
                "verified": True,
            },
        )

        try:
            credentials.oauth
        except XeroNotVerified:
            self.fail("Credentials should have been verified")

    @patch("requests.post")
    def test_verify_failure(self, r_post):
        "If verification credentials are bad, an error is raised"
        r_post.return_value = Mock(
            status_code=401,
            text="oauth_problem=bad_verifier&oauth_problem_advice=The consumer was denied access to this resource.",
        )

        credentials = PublicCredentials(
            consumer_key="key",
            consumer_secret="secret",
            oauth_token="token",
            oauth_token_secret="token_secret",
        )

        with self.assertRaises(XeroUnauthorized):
            credentials.verify("badverifier")

        with self.assertRaises(XeroNotVerified):
            credentials.oauth

    def test_expired(self):
        "Expired credentials are correctly detected"
        now = datetime(2014, 1, 1, 12, 0, 0)
        soon = now + timedelta(minutes=30)

        credentials = PublicCredentials(
            consumer_key="key",
            consumer_secret="secret",
            oauth_token="token",
            oauth_token_secret="token_secret",
        )

        # At this point, oauth_expires_at isn't set
        with self.assertRaises(XeroException):
            credentials.expired(now)

        # Not yet expired
        credentials.oauth_expires_at = soon
        self.assertFalse(credentials.expired(now=now))

        # Expired
        self.assertTrue(credentials.expired(now=soon))


class PartnerCredentialsTest(unittest.TestCase):
    @patch("requests.post")
    def test_initial_constructor(self, r_post):
        "Initial construction causes a request to get a request token"
        r_post.return_value = Mock(
            status_code=200, text="oauth_token=token&oauth_token_secret=token_secret"
        )

        credentials = PartnerCredentials(
            consumer_key="key",
            consumer_secret="secret",
            rsa_key="abc",
            scope="payroll.endpoint",
        )

        # A HTTP request was made
        self.assertTrue(r_post.called)

        state = credentials.state

        # Expiry times should be calculated
        self.assertIsNotNone(state.pop("oauth_authorization_expires_at"))
        self.assertIsNotNone(state.pop("oauth_expires_at"))

        self.assertEqual(
            state,
            {
                "consumer_key": "key",
                "consumer_secret": "secret",
                "oauth_token": "token",
                "oauth_token_secret": "token_secret",
                "verified": False,
                "scope": "payroll.endpoint",
            },
        )

    @patch("requests.post")
    def test_refresh(self, r_post):
        "Refresh function gets a new token"
        r_post.return_value = Mock(
            status_code=200,
            text="oauth_token=token2&oauth_token_secret=token_secret2&oauth_session_handle=session",
        )

        credentials = PartnerCredentials(
            consumer_key="key",
            consumer_secret="secret",
            rsa_key="key",
            oauth_token="token",
            oauth_token_secret="token_secret",
            verified=True,
        )

        credentials.refresh()

        # Expiry times should be calculated
        state = credentials.state
        self.assertIsNotNone(state.pop("oauth_authorization_expires_at"))
        self.assertIsNotNone(state.pop("oauth_expires_at"))

        self.assertEqual(
            state,
            {
                "consumer_key": "key",
                "consumer_secret": "secret",
                "oauth_token": "token2",
                "oauth_token_secret": "token_secret2",
                "oauth_session_handle": "session",
                "verified": True,
            },
        )

    @patch("requests.post")
    def test_configurable_url(self, r_post):
        "Test configurable API url"
        r_post.return_value = Mock(
            status_code=200,
            text="oauth_token=token&oauth_token_secret=token_secret&oauth_session_handle=session",
        )

        url = "https//api-tls.xero.com"

        credentials = PartnerCredentials(
            consumer_key="key",
            consumer_secret="secret",
            rsa_key="key",
            oauth_token="token",
            oauth_token_secret="token_secret",
            verified=True,
            api_url=url,
        )

        credentials.refresh()

        self.assertEqual(
            credentials.url, "{url}/oauth/Authorize?oauth_token=token".format(url=url)
        )


class PrivateCredentialsTest(unittest.TestCase):
    def test_default_url(self):
        "Test default API url"

        credentials = PrivateCredentials(consumer_key="key", rsa_key="rsa_key")

        self.assertEqual(credentials.base_url, "https://api.xero.com")

    def test_configurable_url(self):
        "Test configurable API url"
        url = "https//api-tls.xero.com"

        credentials = PrivateCredentials(
            consumer_key="key", rsa_key="rsa_key", api_url=url
        )

        self.assertEqual(credentials.base_url, url)


class OAuth2CredentialsTest(unittest.TestCase):
    callback_uri = "https://myapp.example.com/xero/auth/callback/"

    def setUp(self):
        super(OAuth2CredentialsTest, self).setUp()
        # Create an expired token to be used by tests
        self.expired_token = {
            "access_token": "1234567890",
            "expires_in": 1800,
            "token_type": "Bearer",
            "refresh_token": "0987654321",
            # 'expires_at': datetime.utcnow().timestamp()}
            "expires_at": time.time(),
        }

    def test_authorisation_url_and_random_state(self):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", callback_uri=self.callback_uri
        )
        url = credentials.generate_url()
        self.assertTrue(url.startswith(XERO_OAUTH2_AUTHORIZE_URL))
        qs = parse_qs(urlparse(url).query)
        # Test that the credentials object can be dumped by state
        cred_state = credentials.state
        # Then test that the relevant attributes are in the querystring
        self.assertEqual(qs["client_id"][0], cred_state["client_id"])
        self.assertEqual(qs["redirect_uri"][0], cred_state["callback_uri"])
        self.assertEqual(qs["response_type"][0], "code")
        self.assertEqual(qs["scope"][0], " ".join(cred_state["scope"]))
        self.assertEqual(qs["state"][0], cred_state["auth_state"])

    def test_authorisation_url_using_initial_state(self):
        credentials = OAuth2Credentials(
            "client_id",
            "client_secret",
            callback_uri=self.callback_uri,
            auth_state="test_state",
        )
        url = urlparse(credentials.generate_url())
        self.assertEqual(credentials.auth_state, "test_state")
        qs = parse_qs(url.query)
        self.assertEqual(qs["state"][0], "test_state")

    @patch("requests_oauthlib.OAuth2Session.request")
    def test_verification_using_bad_auth_uri(self, r_request):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", auth_state="test_state"
        )
        bad_auth_uri = "{}?error=access_denied&state={}".format(
            self.callback_uri, credentials.auth_state
        )
        with self.assertRaises(XeroAccessDenied):
            credentials.verify(bad_auth_uri)
        with self.assertRaises(XeroAccessDenied):
            OAuth2Credentials(
                "client_id",
                "client_secret",
                auth_state="test_state",
                auth_secret=bad_auth_uri,
            )
        self.assertFalse(r_request.called)

    @patch("requests_oauthlib.OAuth2Session.request")
    def test_verification_success(self, r_request):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", auth_state="test_state"
        )
        auth_uri = "{}?code=0123456789&scope={}&state={}".format(
            self.callback_uri, "%20".join(credentials.scope), credentials.auth_state
        )
        r_request.return_value = Mock(
            status_code=200,
            request=Mock(headers={}, body=""),
            headers={},
            text='{"access_token":"1234567890","expires_in":1800,'
            '"token_type":"Bearer","refresh_token":"0987654321"}',
        )
        credentials.verify(auth_uri)
        self.assertTrue(r_request.called)
        self.assertTrue(credentials.token)
        self.assertTrue(credentials.oauth)
        self.assertFalse(credentials.expired())

        # Finally test the state
        self.assertEqual(
            credentials.state,
            {
                "client_id": credentials.client_id,
                "client_secret": credentials.client_secret,
                "auth_state": credentials.auth_state,
                "scope": credentials.scope,
                "user_agent": credentials.user_agent,
                "token": credentials.token,
            },
        )

    @patch("requests_oauthlib.OAuth2Session.request")
    def test_verification_failure(self, r_request):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", auth_state="test_state"
        )
        auth_uri = "{}?code=0123456789&scope={}&state={}".format(
            self.callback_uri, "%20".join(credentials.scope), credentials.auth_state
        )
        r_request.return_value = Mock(
            status_code=400,
            request=Mock(headers={}, body=""),
            headers={},
            text='{"error":"invalid_grant"}',
        )
        with self.assertRaises(XeroAccessDenied):
            credentials.verify(auth_uri)
        with self.assertRaises(XeroAccessDenied):
            OAuth2Credentials(
                "client_id",
                "client_secret",
                auth_state="test_state",
                auth_secret=auth_uri,
            )

    @patch("requests_oauthlib.OAuth2Session.post")
    def test_token_refresh(self, r_post):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", token=self.expired_token
        )
        self.assertTrue(credentials.oauth)
        self.assertTrue(credentials.expired())

        r_post.return_value = Mock(
            status_code=200,
            headers={},
            text='{"access_token":"5555555555","expires_in":1800,'
            '"token_type":"Bearer","refresh_token":"44444444444"}',
        )
        credentials.refresh()
        self.assertTrue(r_post.called)
        self.assertFalse(credentials.expired())
        # Test that the headers were set correctly
        auth = r_post.call_args[1]["auth"]
        self.assertEqual(auth.username, "client_id")
        self.assertEqual(auth.password, "client_secret")

    @patch("requests.get")
    def test_get_tenants(self, r_get):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", token=self.expired_token
        )

        content = '[{"id":"1","tenantId":"12345","tenantType":"ORGANISATION"}]'

        def json_fct():
            return json.loads(content)

        r_get.return_value = Mock(status_code=200, json=json_fct)
        tenants = credentials.get_tenants()
        self.assertTrue(r_get.called)
        self.assertEqual(
            tenants, [{"id": "1", "tenantId": "12345", "tenantType": "ORGANISATION"}]
        )

    @patch("xero.auth.OAuth2Credentials.get_tenants")
    def test_set_default_tenant(self, get_tenants):
        get_tenants.return_value = [
            {"id": "1", "tenantId": "12345", "tenantType": "ORGANISATION"}
        ]
        credentials = OAuth2Credentials(
            "client_id", "client_secret", token=self.expired_token
        )
        credentials.set_default_tenant()
        self.assertEqual(credentials.tenant_id, "12345")

    @patch("requests.get")
    def test_tenant_is_used_in_xero_request(self, r_get):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", token=self.expired_token, tenant_id="12345"
        )
        xero = Xero(credentials)
        # Just return any old response
        r_get.return_value = None
        try:
            xero.contacts.all()
        except:  # NOQA: E722
            pass
        self.assertEqual(r_get.call_args[1]["headers"]["Xero-tenant-id"], "12345")

    def test_tenant_id_not_set_raises_error(self):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", token=self.expired_token
        )
        xero = Xero(credentials)
        with self.assertRaises(XeroTenantIdNotSet):
            xero.contacts.all()

    @patch.object(OAuth2Credentials, "get_tenants", Mock(return_value=[]))
    def test_set_default_tenant_raises_exception(self):
        credentials = OAuth2Credentials(
            "client_id", "client_secret", token=self.expired_token
        )
        with self.assertRaises(XeroException):
            credentials.set_default_tenant()
